# General modules
import numpy as np

# GPAW modules
from gpaw.response.frequencies import ComplexFrequencyDescriptor
from gpaw.response.chiks import ChiKSCalculator
from gpaw.response.localft import LocalFTCalculator, add_LSDA_Bxc
from gpaw.response.site_kernels import SiteKernels

# ASE modules
from ase.units import Hartree


class IsotropicExchangeCalculator:
    r"""Calculator class for the Heisenberg exchange constants

    _           2
    J^ab(q) = - ‾‾ B^(xc†) K^(a†)(q) χ_KS^('+-)(q) K^b(q) B^(xc)
                V0

    calculated for an isotropic system in a plane wave representation using
    the magnetic force theorem within second order perturbation theory, see
    [arXiv:2204.04169].

    Entering the formula for the isotropic exchange constant at wave vector q
    between sublattice a and b is the unit cell volume V0, the functional
    derivative of the (LDA) exchange-correlation energy with respect to the
    magnitude of the magnetization B^(xc), the sublattice site kernels K^a(q)
    and K^b(q) as well as the reactive part of the static transverse magnetic
    susceptibility of the Kohn-Sham system χ_KS^('+-)(q).

    The site kernels encode the partitioning of real space into sites of the
    Heisenberg model. This is not a uniquely defined procedure, why the user
    has to define them externally through the SiteKernels interface."""

    def __init__(self,
                 chiks_calc: ChiKSCalculator,
                 localft_calc: LocalFTCalculator):
        """Construct the IsotropicExchangeCalculator object."""
        # Check that chiks has the assumed properties
        assumed_props = dict(
            gammacentered=True,
            nblocks=1
        )
        for key, item in assumed_props.items():
            assert getattr(chiks_calc, key) == item,\
                f'Expected chiks.{key} == {item}. '\
                f'Got: {getattr(chiks_calc, key)}'

        self.chiks_calc = chiks_calc
        self.context = chiks_calc.context

        # Check assumed properties of the LocalFTCalculator
        assert localft_calc.context is self.context
        assert localft_calc.gs is chiks_calc.gs
        self.localft_calc = localft_calc

        # Bxc field buffer
        self._Bxc_G = None

        # chiksr buffer
        self._chiksr = None

    def __call__(self, q_c, site_kernels: SiteKernels, txt=None):
        """Calculate the isotropic exchange constants for a given wavevector.

        Parameters
        ----------
        q_c : nd.array
            Wave vector q in relative coordinates
        site_kernels : SiteKernels
            Site kernels instance defining the magnetic sites of the crystal
        txt : str
            Separate file to store the chiks calculation output in (optional).
            If not supplied, the output will be written to the standard text
            output location specified when initializing chiks.

        Returns
        -------
        J_abp : nd.array (dtype=complex)
            Isotropic Heisenberg exchange constants between magnetic sites a
            and b for all the site partitions p given by the site_kernels.
        """
        # Get ingredients
        Bxc_G = self.get_Bxc()
        chiksr = self.get_chiksr(q_c, txt=txt)
        qpd, chiksr_GG = chiksr.qpd, chiksr.array[0]  # array = chiksr_zGG
        V0 = qpd.gd.volume

        # Allocate an array for the exchange constants
        nsites = site_kernels.nsites
        J_pab = np.empty(site_kernels.shape + (nsites,), dtype=complex)

        # Compute exchange coupling
        for J_ab, K_aGG in zip(J_pab, site_kernels.calculate(qpd)):
            for a in range(nsites):
                for b in range(nsites):
                    J = np.conj(Bxc_G) @ np.conj(K_aGG[a]).T @ chiksr_GG \
                        @ K_aGG[b] @ Bxc_G
                    J_ab[a, b] = - 2. * J / V0

        # Transpose to have the partitions index last
        J_abp = np.transpose(J_pab, (1, 2, 0))

        return J_abp * Hartree  # Convert from Hartree to eV

    def get_Bxc(self):
        """Get B^(xc)_G from buffer."""
        if self._Bxc_G is None:  # Calculate if buffer is empty
            self._Bxc_G = self._calculate_Bxc()

        return self._Bxc_G

    def _calculate_Bxc(self):
        """Use the PlaneWaveBxc calculator to calculate the plane wave
        coefficients B^xc_G"""
        # Create a plane wave descriptor encoding the plane wave basis. Input
        # q_c is arbitrary, since we are assuming that chiks.gammacentered == 1
        qpd0 = self.chiks_calc.get_pw_descriptor([0., 0., 0.])

        return self.localft_calc(qpd0, add_LSDA_Bxc)

    def get_chiksr(self, q_c, txt=None):
        """Get χ_KS^('+-)(q) from buffer."""
        q_c = np.asarray(q_c)

        # Calculate if buffer is empty or a new q-point is given
        if self._chiksr is None or not np.allclose(q_c, self._chiksr.q_c):
            self._chiksr = self._calculate_chiksr(q_c, txt=txt)

        return self._chiksr

    def _calculate_chiksr(self, q_c, txt=None):
        r"""Use the ChiKSCalculator to calculate the reactive part of the
        static Kohn-Sham susceptibility χ_KS^('+-)(q).

        First, the dynamic Kohn-Sham susceptibility

                                 __  __
                              1  \   \        f_nk↑ - f_mk+q↓
        χ_KS,GG'^+-(q,ω+iη) = ‾  /   /  ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                              V  ‾‾  ‾‾ ħω - (ε_mk+q↓ - ε_nk↑) + iħη
                                 k  n,m
                                        x n_nk↑,mk+q↓(G+q) n_mk+q↓,nk↑(-G'-q)

        is calculated in the static limit ω=0 and without broadening η=0. Then,
        the reactive part (see [PRB 103, 245110 (2021)]) is extracted:

                              1
        χ_KS,GG'^(+-')(q,z) = ‾ [χ_KS,GG'^+-(q,z) + χ_KS,-G'-G^-+(-q,-z*)].
                              2
        """
        # Initiate new output file, if supplied
        if txt is not None:
            self.context.new_txt_and_timer(txt)

        zd = ComplexFrequencyDescriptor.from_array([0. + 0.j])
        chiks = self.chiks_calc.calculate('+-', q_c, zd)
        if np.allclose(q_c, 0.):
            chiks.symmetrize_reciprocity()

        # Take the reactive part
        chiksr = chiks.copy_reactive_part()

        return chiksr



from gpaw.test import equal
from gpaw.grid_descriptor import GridDescriptor
from gpaw.spline import Spline
import gpaw.mpi as mpi
from gpaw.lfc import LocalizedFunctionsCollection as LFC


def test_lfc_lf():
    s = Spline(0, 1.0, [1.0, 0.5, 0.0])
    n = 40
    a = 8.0
    gd = GridDescriptor((n, n, n), (a, a, a), comm=mpi.serial_comm)
    c = LFC(gd, [[s], [s], [s]])
    c.set_positions([(0.5, 0.5, 0.25 + 0.25 * i) for i in [0, 1, 2]])
    b = gd.zeros()
    c.add(b)
    x = gd.integrate(b)

    gd = GridDescriptor((n, n, n), (a, a, a), comm=mpi.serial_comm)
    c = LFC(gd, [[s], [s], [s]])
    c.set_positions([(0.5, 0.5, 0.25 + 0.25 * i) for i in [0, 1, 2]])
    b = gd.zeros()
    c.add(b)
    y = gd.integrate(b)
    equal(x, y, 1e-13)



"""This module is used to generate atomic orbital basis sets."""
import sys
from io import StringIO

import numpy as np
from ase.units import Hartree
from gpaw.utilities import devnull

from gpaw import __version__ as version
from gpaw.utilities import divrl
from gpaw.atom.generator import Generator
from gpaw.atom.all_electron import AllElectron
from gpaw.atom.configurations import parameters
from gpaw.basis_data import Basis, BasisFunction, get_basis_name
from gpaw.atom.radialgd import (AERadialGridDescriptor,
                                EquidistantRadialGridDescriptor)


def get_basis_l(f_j, l_j, Nv):
    N = 0
    lvalues = []
    j_l = []
    Nl_j = len(l_j)
    while N < Nv:
        Nl_j -= 1
        if f_j[Nl_j] > 0:
            lvalues.append(l_j[Nl_j])
            N += f_j[Nl_j]
            j_l.append(Nl_j)
    assert N == Nv

    if min(lvalues) != 0:  # always include s-orbital !
        lvalues = lvalues + [0]
        reversed_l_j = list(l_j)
        reversed_l_j.reverse()
        j = len(reversed_l_j) - reversed_l_j.index(0) - 1
        j_l.append(j)

    return [j_l, lvalues]


def get_gaussianlike_basis_function(rgd, l, rchar, gcut):
    rcut = rgd.r_g[gcut]
    gaussian = QuasiGaussian(1.0 / rchar**2, rcut)
    r_g = rgd.r_g[:gcut + 1]
    phit_g = gaussian(r_g) * r_g**l
    phit_g[-1] = 0.0
    norm = np.dot(rgd.dr_g[:gcut + 1], (r_g * phit_g)**2)**0.5
    phit_g /= norm
    norm2 = np.dot(rgd.dr_g[:gcut + 1], (r_g * phit_g)**2)**0.5
    assert abs(1.0 - norm2) < 1e-10
    return phit_g


def make_split_valence_basis_function(r_g, psi_g, l, gcut):
    """Get polynomial which joins psi smoothly at rcut.

    Returns an array of function values f(r) * r, where::

              l           2
      f(r) = r  * (a - b r ),  r < rcut
      f(r) = psi(r),           r >= rcut

    where a and b are determined such that f(r) is continuous and
    differentiable at rcut.  The parameter psi should be an atomic
    orbital.
    """
    r1 = r_g[gcut]  # ensure that rcut is moved to a grid point
    r2 = r_g[gcut + 1]
    y1 = psi_g[gcut] / r_g[gcut]
    y2 = psi_g[gcut + 1] / r_g[gcut + 1]
    b = - (y2 / r2**l - y1 / r1**l) / (r2**2 - r1**2)
    a = (y1 / r1**l + b * r1**2)
    psi_g2 = r_g**(l + 1) * (a - b * r_g**2)
    psi_g2[gcut:] = psi_g[gcut:]
    return psi_g2


def rsplit_by_norm(rgd, l, u, tailnorm_squared, txt):
    """Find radius outside which remaining tail has a particular norm."""
    norm_squared = np.dot(rgd.dr_g, u * u)
    partial_norm_squared = 0.
    i = len(u) - 1
    absolute_tailnorm_squared = tailnorm_squared * norm_squared
    while partial_norm_squared < absolute_tailnorm_squared:
        # Integrate backwards.  This is important since the pseudo
        # wave functions have strange behaviour near the core.
        partial_norm_squared += rgd.dr_g[i] * u[i]**2
        i -= 1
    rsplit = rgd.r_g[i + 1]
    msg = ('Tail norm %.03f :: rsplit=%.02f Bohr' %
           ((partial_norm_squared / norm_squared)**0.5, rsplit))
    print(msg, file=txt)
    gsplit = rgd.floor(rsplit)
    splitwave = make_split_valence_basis_function(rgd.r_g, u, l, gsplit)
    return rsplit, partial_norm_squared, splitwave


class BasisMaker:
    """Class for creating atomic basis functions."""
    def __init__(self, generator, name=None, run=True, gtxt='-',
                 non_relativistic_guess=False, xc='PBE',
                 save_setup=False):

        if isinstance(generator, str):  # treat 'generator' as symbol
            generator = Generator(generator, scalarrel=True,
                                  xcname=xc, txt=gtxt,
                                  nofiles=True)
            generator.N *= 4
        self.generator = generator
        self.rgd = AERadialGridDescriptor(generator.beta / generator.N,
                                          1.0 / generator.N, generator.N,
                                          default_spline_points=100)
        self.name = name

        if run:
            if non_relativistic_guess:
                ae0 = AllElectron(generator.symbol, scalarrel=False,
                                  nofiles=False, txt=gtxt, xcname=xc)
                ae0.N = generator.N
                ae0.beta = generator.beta
                ae0.run()
                # Now files will be stored such that they can
                # automagically be used by the next run()
            setup = generator.run(write_xml=False,
                                  name=name,
                                  **parameters[generator.symbol])

            if save_setup:
                setup.write_xml()
        else:
            if save_setup:
                raise ValueError('cannot save setup here because setup '
                                 'was already generated before basis '
                                 'generation.')

    def smoothify(self, psi_mg, l):
        r"""Generate pseudo wave functions from all-electron ones.

        The pseudo wave function is::

                                   ___
               ~                   \    /   ~             \    ~     ~
            | psi  > = | psi  > +   )  | | phi > - | phi > ) < p  | psi > ,
                 m          m      /__  \     i         i /     i      m
                                    i

        where the scalar products are found by solving::

                            ___
              ~             \     ~             ~     ~
            < p | psi  > =   )  < p  | phi  > < p  | psi  > .
               i     m      /__    i      j      j      m
                             j

        In order to ensure smoothness close to the core, the
        all-electron wave function and partial wave are then
        multiplied by a radial function which approaches 0 near the
        core, such that the pseudo wave function approaches::

                        ___
               ~        \      ~       ~     ~
            | psi  > =   )  | phi >  < p  | psi >    (for r << rcut),
                 m      /__      i      i      m
                         i

        which is exact if the projectors/pseudo partial waves are complete.
        """
        if psi_mg.ndim == 1:
            return self.smoothify(psi_mg[None], l)[0]

        g = self.generator
        u_ng = g.u_ln[l]
        q_ng = g.q_ln[l]
        s_ng = g.s_ln[l]

        Pi_nn = np.dot(g.r * q_ng, u_ng.T)
        Q_nm = np.dot(g.r * q_ng, psi_mg.T)
        Qt_nm = np.linalg.solve(Pi_nn, Q_nm)

        # Weight-function for truncating all-electron parts smoothly near core
        gmerge = g.r2g(g.rcut_l[l])
        w_g = np.ones(g.r.shape)
        w_g[0:gmerge] = (g.r[0:gmerge] / g.r[gmerge])**2.
        w_g = w_g[None]

        psit_mg = psi_mg * w_g + np.dot(Qt_nm.T, s_ng - u_ng * w_g)
        return psit_mg

    def make_orbital_vector(self, j, rcut, vconf=None):
        """Returns a smooth basis vector given an all-electron one."""
        l = self.generator.l_j[j]
        psi_g, e = self.generator.solve_confined(j, rcut, vconf)
        psit_g = self.smoothify(psi_g, l)
        return psit_g

    def rcut_by_energy(self, j, esplit=.1, tolerance=.1, rguess=6.,
                       vconf_args=None):
        """Find confinement cutoff corresponding to given orbital energy shift.

        Creates a confinement potential for the orbital given by j,
        such that the confined-orbital energy is (emin to emax) eV larger
        than the free-orbital energy."""
        g = self.generator
        e_base = g.e_j[j]
        rc = rguess

        if vconf_args is None:
            vconf = None
        else:
            amplitude, ri_rel = vconf_args
            vconf = g.get_confinement_potential(amplitude, ri_rel * rc, rc)

        psi_g, e = g.solve_confined(j, rc, vconf)
        de_min, de_max = esplit / Hartree, (esplit + tolerance) / Hartree

        rmin = 0.
        rmax = g.r[-1]

        de = e - e_base
        # print '--------'
        # print 'Start bisection'
        # print'e_base =',e_base
        # print 'e =',e
        # print '--------'
        while de < de_min or de > de_max:
            if de < de_min:  # Move rc left -> smaller cutoff, higher energy
                rmax = rc
                rc = (rc + rmin) / 2.
            else:  # Move rc right
                rmin = rc
                rc = (rc + rmax) / 2.
            if vconf is not None:
                vconf = g.get_confinement_potential(amplitude, ri_rel * rc, rc)
            psi_g, e = g.solve_confined(j, rc, vconf)
            de = e - e_base
            # print 'rc = %.03f :: e = %.03f :: de = %.03f' % (rc, e*Hartree,
            #                                                  de*Hartree)
            # if rmin - rmax < 1e-
            if g.r2g(rmax) - g.r2g(rmin) <= 1:  # adjacent points
                break  # cannot meet tolerance due to grid resolution
        # print 'Done!'
        return psi_g, e, de, vconf, rc

    def generate(self, zetacount=2, polarizationcount=1,
                 tailnorm=(0.16, 0.3, 0.6), energysplit=0.1, tolerance=1.0e-3,
                 referencefile=None, referenceindex=None, rcutpol_rel=1.0,
                 rcutmax=20.0,
                 rcharpol_rel=None,
                 vconf_args=(12.0, 0.6), txt='-',
                 include_energy_derivatives=False,
                 # lvalues=None, # XXX clean up some of these!
                 jvalues=None,
                 l_pol=None):
        """Generate an entire basis set.

        This is a high-level method which will return a basis set
        consisting of several different basis vector types.

        Parameters:

        ===================== =================================================
        ``zetacount``         Number of basis functions per occupied orbital
        ``polarizationcount`` Number of polarization functions
        ``tailnorm``          List of tail norms for split-valence scheme
        ``energysplit``       Energy increase defining confinement radius (eV)
        ``tolerance``         Tolerance of energy split (eV)
        ``referencefile``     gpw-file used to generate polarization function
        ``referenceindex``    Index in reference system of relevant atom
        ``rcutpol_rel``       Polarization rcut relative to largest other rcut
        ``rcutmax``           No cutoff will be greater than this value
        ``vconf_args``        Parameters (alpha, ri/rc) for conf. potential
        ``txt``               Log filename or '-' for stdout
        ===================== =================================================

        Returns a fully initialized Basis object.
        """

        if txt == '-':
            txt = sys.stdout
        elif txt is None:
            txt = devnull

        if isinstance(tailnorm, float):
            tailnorm = (tailnorm,)
        if 1 + len(tailnorm) < max(polarizationcount, zetacount):
            raise ValueError(
                'Needs %d tail norm values, but only %d are specified' %
                (max(polarizationcount, zetacount) - 1, len(tailnorm)))

        textbuffer = StringIO()

        class TeeStream:  # quick hack to both write and save output
            def __init__(self, out1, out2):
                self.out1 = out1
                self.out2 = out2

            def write(self, string):
                self.out1.write(string)
                self.out2.write(string)

        txt = TeeStream(txt, textbuffer)

        if vconf_args is not None:
            amplitude, ri_rel = vconf_args

        g = self.generator
        rgd = self.rgd

        njcore = g.njcore
        n_j = g.n_j[njcore:]
        l_j = g.l_j[njcore:]
        f_j = g.f_j[njcore:]

        if jvalues is None:
            jvalues = []
            sortkeys = []
            for j in range(len(n_j)):
                if f_j[j] == 0 and l_j[j] != 0:
                    continue
                jvalues.append(j)
                sortkeys.append(l_j[j])

            # Now order jvalues by l
            #
            # Use a stable sort so the energy ordering within each
            # angular momentum is guaranteed to be preserved
            args = np.argsort(sortkeys, kind='mergesort')
            jvalues = np.array(jvalues)[args]

        fulljvalues = [njcore + j for j in jvalues]

        if isinstance(energysplit, float):
            energysplit = [energysplit] * len(jvalues)

        title = '%s Basis functions for %s' % (g.xcname, g.symbol)
        print(title, file=txt)
        print('=' * len(title), file=txt)

        singlezetas = []
        energy_derivative_functions = []
        multizetas = [[] for i in range(zetacount - 1)]
        polarization_functions = []

        splitvalencedescr = 'split-valence wave, fixed tail norm'
        derivativedescr = 'derivative of sz wrt. (ri/rc) of potential'

        for vj, fullj, esplit in zip(jvalues, fulljvalues, energysplit):
            l = l_j[vj]
            n = n_j[vj]
            assert n > 0
            orbitaltype = str(n) + 'spdf'[l]
            msg = 'Basis functions for l=%d, n=%d' % (l, n)
            print(file=txt)
            print(msg + '\n', '-' * len(msg), file=txt)
            print(file=txt)
            if vconf_args is None:
                adverb = 'sharply'
            else:
                adverb = 'softly'
            print('Zeta 1: %s confined pseudo wave,' % adverb, end=' ',
                  file=txt)

            u, e, de, vconf, rc = self.rcut_by_energy(fullj, esplit,
                                                      tolerance,
                                                      vconf_args=vconf_args)
            if rc > rcutmax:
                rc = rcutmax  # scale things down
                if vconf is not None:
                    vconf = g.get_confinement_potential(amplitude, ri_rel * rc,
                                                        rc)
                u, e = g.solve_confined(fullj, rc, vconf)
                print('using maximum cutoff', file=txt)
                print('rc=%.02f Bohr' % rc, file=txt)
            else:
                print('fixed energy shift', file=txt)
                print('DE=%.03f eV :: rc=%.02f Bohr'
                      % (de * Hartree, rc), file=txt)

            if vconf is not None:
                print('Potential amp=%.02f :: ri/rc=%.02f' %
                      (amplitude, ri_rel), file=txt)
            phit_g = self.smoothify(u, l)
            bf = BasisFunction(n, l, rc, phit_g,
                               '%s-sz confined orbital' % orbitaltype)
            norm = np.dot(g.dr, phit_g * phit_g)**.5
            print('Norm=%.03f' % norm, file=txt)
            singlezetas.append(bf)

            zetacounter = iter(range(2, zetacount + 1))

            if include_energy_derivatives:
                assert zetacount > 1
                zeta = next(zetacounter)
                print('\nZeta %d: %s' % (zeta, derivativedescr), file=txt)
                vconf2 = g.get_confinement_potential(amplitude,
                                                     ri_rel * rc * .99, rc)
                u2, e2 = g.solve_confined(fullj, rc, vconf2)

                phit2_g = self.smoothify(u2, l)
                dphit_g = phit2_g - phit_g
                dphit_norm = np.dot(rgd.dr_g, dphit_g * dphit_g) ** .5
                dphit_g /= dphit_norm
                descr = '%s-dz E-derivative of sz' % orbitaltype
                bf = BasisFunction(None, l, rc, dphit_g, descr)
                energy_derivative_functions.append(bf)

            for i, zeta in enumerate(zetacounter):
                print('\nZeta %d: %s' % (zeta, splitvalencedescr), file=txt)
                # Unresolved issue:  how does the lack of normalization
                # of the first function impact the tail norm scheme?
                # Presumably not much, since most interesting stuff happens
                # close to the core.
                rsplit, norm, splitwave = rsplit_by_norm(rgd, l, phit_g,
                                                         tailnorm[i]**2.0,
                                                         txt)
                descr = '%s-%sz split-valence wave' % (orbitaltype,
                                                       '0sdtq56789'[zeta])
                bf = BasisFunction(None, l, rsplit, phit_g - splitwave, descr)
                multizetas[i].append(bf)

        if polarizationcount > 0 or l_pol is not None:
            if l_pol is None:
                # Now make up some properties for the polarization orbital
                # We just use the cutoffs from the previous one times a factor
                # Find 'missing' values in lvalues
                lvalues = [l_j[vj] for vj in jvalues]
                for i in range(max(lvalues) + 1):
                    if list(lvalues).count(i) == 0:
                        l_pol = i
                        break
                else:
                    l_pol = max(lvalues) + 1

            # Find the last state with l=l_pol - 1, which will be the state we
            # base the polarization function on
            for vj, fullj, bf in zip(jvalues[::-1], fulljvalues[::-1],
                                     singlezetas[::-1]):
                if bf.l == l_pol - 1:
                    fullj_pol = fullj
                    rcut = bf.rc * rcutpol_rel
                    break
            else:
                raise ValueError('The requested value l_pol=%d requires l=%d '
                                 'among valence states' % (l_pol, l_pol - 1))
            rcut = min(rcut, rcutmax)
            msg = 'Polarization function: l=%d, rc=%.02f' % (l_pol, rcut)
            print('\n' + msg, file=txt)
            print('-' * len(msg), file=txt)
            # Make a single Gaussian for polarization function.
            #
            # It is known that for given l, the sz cutoff defined
            # by some fixed energy is strongly correlated to the
            # value of the characteristic radius which best reproduces
            # the wave function found by interpolation.
            #
            # We know that for e.g. d orbitals:
            #   rchar ~= .37 rcut[sz](.3eV)
            # Since we don't want to spend a lot of time finding
            # these value for other energies, we just find the energy
            # shift at .3 eV now

            u, e, de, vconf, rc_fixed = self.rcut_by_energy(fullj_pol,
                                                            .3, 1e-2,
                                                            6., (12., .6))

            default_rchar_rel = .25
            # Defaults for each l.  Actually we don't care right now
            rchar_rels = {}

            if rcharpol_rel is None:
                rcharpol_rel = rchar_rels.get(l_pol, default_rchar_rel)
            rchar = rcharpol_rel * rc_fixed
            gaussian = QuasiGaussian(1.0 / rchar**2, rcut)
            psi_pol = gaussian(rgd.r_g) * rgd.r_g**(l_pol + 1)
            norm = np.dot(rgd.dr_g, psi_pol * psi_pol) ** .5
            psi_pol /= norm
            print('Single quasi Gaussian', file=txt)
            msg = 'Rchar = %.03f*rcut = %.03f Bohr' % (rcharpol_rel, rchar)
            adjective = 'Gaussian'
            print(msg, file=txt)
            type = '%s-type %s polarization' % ('spdfg'[l_pol], adjective)
            bf_pol = BasisFunction(None, l_pol, rcut, psi_pol, type)

            polarization_functions.append(bf_pol)
            for i in range(polarizationcount - 1):
                npol = i + 2
                msg = '\n%s: %s' % (['Secondary', 'Tertiary', 'Quaternary',
                                     'Quintary', 'Sextary', 'Septenary'][i],
                                    splitvalencedescr)
                print(msg, file=txt)
                rsplit, norm, splitwave = rsplit_by_norm(rgd, l_pol, psi_pol,
                                                         tailnorm[i],
                                                         txt)
                descr = ('%s-type split-valence polarization %d'
                         % ('spdfg'[l_pol], npol))
                bf_pol = BasisFunction(None, l_pol, rsplit,
                                       psi_pol - splitwave,
                                       descr)
                polarization_functions.append(bf_pol)

        bf_j = []
        bf_j.extend(singlezetas)
        bf_j.extend(energy_derivative_functions)
        for multizeta_list in multizetas:
            bf_j.extend(multizeta_list)
        bf_j.extend(polarization_functions)

        rcmax = max([bf.rc for bf in bf_j])

        # The non-equidistant grids are really only suited for AE WFs
        d = 1.0 / 64
        equidistant_grid = np.arange(0.0, rcmax + d, d)
        ng = len(equidistant_grid)

        for bf in bf_j:
            # We have been storing phit_g * r, but we just want phit_g
            bf.phit_g = divrl(bf.phit_g, 1, rgd.r_g)

            gcut = min(int(1 + bf.rc / d), ng - 1)

            assert equidistant_grid[gcut] >= bf.rc
            assert equidistant_grid[gcut - 1] <= bf.rc

            bf.rc = equidistant_grid[gcut]
            # Note: bf.rc *must* correspond to a grid point (spline issues)
            bf.ng = gcut + 1
            # XXX all this should be done while building the basis vectors,
            # not here

            # Quick hack to change to equidistant coordinates
            spline = rgd.spline(bf.phit_g, rgd.r_g[rgd.floor(bf.rc)], bf.l,
                                points=100)
            bf.phit_g = np.array([spline(r) * r**bf.l
                                  for r in equidistant_grid[:bf.ng]])
            bf.phit_g[-1] = 0.

        basistype = get_basis_name(zetacount, polarizationcount)
        if self.name is None:
            compound_name = basistype
        else:
            compound_name = '%s.%s' % (self.name, basistype)

        basis = Basis(g.symbol, compound_name, False,
                      EquidistantRadialGridDescriptor(d, ng))
        basis.bf_j = bf_j
        basis.generatordata = textbuffer.getvalue().strip()
        basis.generatorattrs = {'version': version}
        textbuffer.close()

        return basis

    def grplot(self, bf_j):
        """Plot basis functions on generator's radial grid."""
        import matplotlib.pyplot as plt
        rc = max([bf.rc for bf in bf_j])
        g = self.generator
        r = g.r
        for bf in bf_j:
            label = bf.type
            # XXX times g.r or not times g.r ?
            plt.plot(r, bf.phit_g / r, label=label[:12])
        axis = plt.axis()
        newaxis = [0., rc, axis[2], axis[3]]
        plt.axis(newaxis)
        plt.legend()
        plt.show()

    def plot(self, basis, figure=None, title=None, filename=None):
        """Plot basis functions using MatPlotLib."""
        # XXX method should no longer belong to a basis maker
        import matplotlib.pyplot as plt
        rc = max([bf.rc for bf in basis.bf_j])
        r = np.linspace(0., basis.d * (basis.ng - 1), basis.ng)
        g = self.generator
        if figure is not None:
            plt.figure(figure)
        else:
            plt.figure()  # not very elegant
        if title is None:
            title = g.symbol
        plt.title(title)
        for bf in basis.bf_j:
            label = bf.type
            # XXX times g.r or not times g.r ?
            phit_g = np.zeros_like(r)
            phit_g[:len(bf.phit_g)] = bf.phit_g
            plt.plot(r, phit_g * r, label=label[:12])
        axis = plt.axis()
        newaxis = [0., rc, axis[2], axis[3]]
        plt.axis(newaxis)
        plt.legend()
        if filename is not None:
            plt.savefig(filename)


class QuasiGaussian:
    """Gaussian-like functions for expansion of orbitals.

    Implements f(r) = A [G(r) - P(r)] where::

      G(r) = exp{- alpha r^2}
      P(r) = a - b r^2

    with (a, b) such that f(rcut) == f'(rcut) == 0.
    """
    def __init__(self, alpha, rcut, A=1.):
        self.alpha = alpha
        self.rcut = rcut
        expmar2 = np.exp(-alpha * rcut**2)
        a = (1 + alpha * rcut**2) * expmar2
        b = alpha * expmar2
        self.a = a
        self.b = b
        self.A = A

    def __call__(self, r):
        """Evaluate function values at r, which is a numpy array."""
        condition = (r < self.rcut) & (self.alpha * r**2 < 700.)
        r2 = np.where(condition, r**2., 0.)  # prevent overflow
        g = np.exp(-self.alpha * r2)
        p = (self.a - self.b * r2)
        y = np.where(condition, g - p, 0.)
        return self.A * y

    def renormalize(self, norm):
        """Divide function by norm."""
        self.A /= norm



import numpy as np
from numpy.linalg import inv, solve

from ase.utils.timing import timer

from gpaw.lcaotddft.hamiltonian import KickHamiltonian
from gpaw import debug
from gpaw.tddft.units import au_to_as
from gpaw.utilities.scalapack import (pblas_simple_hemm, pblas_simple_gemm,
                                      scalapack_inverse, scalapack_solve,
                                      scalapack_tri2full)


def create_propagator(name, **kwargs):
    if name is None:
        return create_propagator('sicn')
    elif isinstance(name, Propagator):
        return name
    elif isinstance(name, dict):
        kwargs.update(name)
        return create_propagator(**kwargs)
    elif name == 'sicn':
        return SICNPropagator(**kwargs)
    elif name == 'ecn':
        return ECNPropagator(**kwargs)
    elif name.endswith('.ulm'):
        return ReplayPropagator(name, **kwargs)
    else:
        raise ValueError('Unknown propagator: %s' % name)


def equal(a, b, eps=1e-8):
    return abs(a - b) < eps


class Propagator(object):

    def __init__(self):
        object.__init__(self)

    def initialize(self, paw):
        self.timer = paw.timer
        self.log = paw.log

    def kick(self, ext, time):
        raise NotImplementedError()

    def propagate(self, time, time_step):
        raise NotImplementedError()

    def control_paw(self, paw):
        raise NotImplementedError()

    def todict(self):
        raise NotImplementedError()

    def get_description(self):
        return '%s' % self.__class__.__name__


class LCAOPropagator(Propagator):

    def __init__(self):
        Propagator.__init__(self)

    def initialize(self, paw):
        Propagator.initialize(self, paw)
        self.wfs = paw.wfs
        self.density = paw.density
        self.hamiltonian = paw.td_hamiltonian


class ReplayPropagator(LCAOPropagator):

    def __init__(self, filename, update='all'):
        from gpaw.lcaotddft.wfwriter import WaveFunctionReader
        LCAOPropagator.__init__(self)
        self.filename = filename
        self.update_mode = update
        self.reader = WaveFunctionReader(self.filename)
        self.read_index = 1
        self.read_count = len(self.reader)

    def _align_read_index(self, time):
        while self.read_index < self.read_count:
            r = self.reader[self.read_index]
            if equal(r.time, time):
                break
            self.read_index += 1
        if self.read_index == self.read_count:
            raise RuntimeError('Time not found: %f' % time)

    def _read(self):
        reader = self.reader[self.read_index]
        r = reader.wave_functions
        self.wfs.read_wave_functions(r)
        self.wfs.read_occupations(r)
        self.read_index += 1

    def kick(self, ext, time):
        self._align_read_index(time)
        # Check that this is the step after kick
        assert not equal(self.reader[self.read_index].time,
                         self.reader[self.read_index + 1].time)
        self._read()
        self.hamiltonian.update(self.update_mode)

    def propagate(self, time, time_step):
        next_time = time + time_step
        self._align_read_index(next_time)
        self._read()
        self.hamiltonian.update(self.update_mode)
        return next_time

    def control_paw(self, paw):
        # Read the initial state
        index = 1
        r = self.reader[index]
        assert r.action == 'init'
        assert equal(r.time, paw.time)
        self.read_index = index
        self._read()
        index += 1
        # Read the rest
        while index < self.read_count:
            r = self.reader[index]
            if r.action == 'init':
                index += 1
            elif r.action == 'kick':
                assert equal(r.time, paw.time)
                paw.absorption_kick(r.kick_strength)
                assert equal(r.time, paw.time)
                index += 1
            elif r.action == 'propagate':
                # Skip earlier times
                if r.time < paw.time or equal(r.time, paw.time):
                    index += 1
                    continue
                # Count the number of steps with the same time step
                time = paw.time
                time_step = r.time - time
                iterations = 0
                while index < self.read_count:
                    r = self.reader[index]
                    if (r.action != 'propagate' or
                        not equal(r.time - time, time_step)):
                        break
                    iterations += 1
                    time = r.time
                    index += 1
                # Propagate
                paw.propagate(time_step * au_to_as, iterations)
                assert equal(time, paw.time)
            else:
                raise RuntimeError('Unknown action: %s' % r.action)

    def __del__(self):
        self.reader.close()

    def todict(self):
        return {'name': self.filename,
                'update': self.update_mode}

    def get_description(self):
        lines = [self.__class__.__name__]
        lines += ['    File: %s' % (self.filename)]
        lines += ['    Update: %s' % (self.update_mode)]
        return '\n'.join(lines)


class ECNPropagator(LCAOPropagator):

    def __init__(self):
        LCAOPropagator.__init__(self)

    def initialize(self, paw, hamiltonian=None):
        LCAOPropagator.initialize(self, paw)
        if hamiltonian is not None:
            self.hamiltonian = hamiltonian

        ksl = self.wfs.ksl
        using_blacs = ksl.using_blacs
        if using_blacs:
            from gpaw.blacs import Redistributor
            self.log('BLACS Parallelization')

            # Propagator function
            self.propagate_wfs = self.propagate_wfs_blacs

            # Parallel grid descriptors
            grid = ksl.blockgrid
            assert grid.nprow * grid.npcol == ksl.block_comm.size
            self.mm_block_descriptor = ksl.mmdescriptor
            self.Cnm_block_descriptor = grid.new_descriptor(ksl.bd.nbands,
                                                            ksl.nao,
                                                            ksl.blocksize,
                                                            ksl.blocksize)
            self.CnM_unique_descriptor = ksl.nM_unique_descriptor

            # Redistributors
            self.Cnm2nM = Redistributor(ksl.block_comm,
                                        self.Cnm_block_descriptor,
                                        self.CnM_unique_descriptor)
            self.CnM2nm = Redistributor(ksl.block_comm,
                                        self.CnM_unique_descriptor,
                                        self.Cnm_block_descriptor)

            for kpt in self.wfs.kpt_u:
                scalapack_tri2full(self.mm_block_descriptor, kpt.S_MM)
                scalapack_tri2full(self.mm_block_descriptor, kpt.T_MM)

            if self.density.gd.comm.rank != 0:
                # This is a (0, 0) dummy array that is needed for
                # redistributing between nM and nm block descriptor.
                # See propagate_wfs() and also
                # BlacsOrbitalLayouts.calculate_blocked_density_matrix()
                self.dummy_C_nM = \
                    self.CnM_unique_descriptor.zeros(dtype=complex)

        else:
            # Propagator function
            self.propagate_wfs = self.propagate_wfs_numpy

        if debug and using_blacs:
            nao = ksl.nao
            self.MM_descriptor = grid.new_descriptor(nao, nao, nao, nao)
            self.mm2MM = Redistributor(ksl.block_comm,
                                       self.mm_block_descriptor,
                                       self.MM_descriptor)
            self.MM2mm = Redistributor(ksl.block_comm,
                                       self.MM_descriptor,
                                       self.mm_block_descriptor)

    def kick(self, ext, time):
        # Propagate
        get_matrix = self.wfs.eigensolver.calculate_hamiltonian_matrix
        kick_hamiltonian = KickHamiltonian(self.hamiltonian.hamiltonian,
                                           self.density, ext)
        for kpt in self.wfs.kpt_u:
            Vkick_MM = get_matrix(kick_hamiltonian, self.wfs, kpt,
                                  add_kinetic=False, root=-1)
            for i in range(10):
                self.propagate_wfs(kpt.C_nM, kpt.C_nM, kpt.S_MM, Vkick_MM, 0.1)

        # Update Hamiltonian (and density)
        self.hamiltonian.update()

    def propagate(self, time, time_step):
        get_H_MM = self.hamiltonian.get_hamiltonian_matrix
        for kpt in self.wfs.kpt_u:
            H_MM = get_H_MM(kpt, time)
            self.propagate_wfs(kpt.C_nM, kpt.C_nM, kpt.S_MM, H_MM, time_step)
        self.hamiltonian.update()
        return time + time_step

    @timer('Linear solve')
    def propagate_wfs_blacs(self, source_C_nM, target_C_nM, S_mm, H_mm, dt):
        # XXX, Preallocate?
        target_C_nm = self.Cnm_block_descriptor.empty(dtype=complex)
        source_C_nm = self.Cnm_block_descriptor.empty(dtype=complex)

        # C_nM is duplicated over all ranks in gd.comm.
        # Master rank will provide the actual data and other
        # ranks use a dummy array in redistribute().
        if self.density.gd.comm.rank != 0:
            source = self.dummy_C_nM
        else:
            source = source_C_nM
        self.CnM2nm.redistribute(source, source_C_nm)

        # Note: tri2full for S_mm and T_mm is done already in initialize().
        # H_mm seems to be a full matrix as we are working with complex
        # dtype, so no need to do tri2full here again XXX
        # scalapack_tri2full(self.mm_block_descriptor, H_mm)
        SjH_mm = S_mm + (0.5j * dt) * H_mm

        # 1. target = (S - 0.5j*H*dt) * source
        pblas_simple_gemm(self.Cnm_block_descriptor,
                          self.mm_block_descriptor,
                          self.Cnm_block_descriptor,
                          source_C_nm,
                          SjH_mm,
                          target_C_nm,
                          transb='C')

        # 2. target = (S + 0.5j*H*dt)^-1 * target
        scalapack_solve(self.mm_block_descriptor,
                        self.Cnm_block_descriptor,
                        SjH_mm,
                        target_C_nm)

        # C_nM is duplicated over all ranks in gd.comm.
        # Master rank will receive the data and other
        # ranks use a dummy array in redistribute()
        if self.density.gd.comm.rank != 0:
            target = self.dummy_C_nM
        else:
            target = target_C_nM
        self.Cnm2nM.redistribute(target_C_nm, target)

        # Broadcast the new C_nM to all ranks in gd.comm
        self.density.gd.comm.broadcast(target_C_nM, 0)

    @timer('Linear solve')
    def propagate_wfs_numpy(self, source_C_nM, target_C_nM, S_MM, H_MM, dt):
        SjH_MM = S_MM + (0.5j * dt) * H_MM
        target_C_nM[:] = np.dot(source_C_nM, SjH_MM.conj().T)
        target_C_nM[:] = solve(SjH_MM.T, target_C_nM.T).T

    def blacs_mm_to_global(self, H_mm):
        if not debug:
            raise RuntimeError('Use debug mode')
        # Someone could verify that this works and remove the error.
        raise NotImplementedError('Method untested and thus unreliable')
        target = self.MM_descriptor.empty(dtype=complex)
        self.mm2MM.redistribute(H_mm, target)
        self.wfs.world.barrier()
        return target

    def blacs_nm_to_global(self, C_nm):
        # Someone could verify that this works and remove the error.
        raise NotImplementedError('Method untested and thus unreliable')
        target = self.CnM_unique_descriptor.empty(dtype=complex)
        self.Cnm2nM.redistribute(C_nm, target)
        self.wfs.world.barrier()
        return target

    def todict(self):
        return {'name': 'ecn'}


class SICNPropagator(ECNPropagator):

    def __init__(self):
        ECNPropagator.__init__(self)

    def initialize(self, paw):
        ECNPropagator.initialize(self, paw)
        # Allocate kpt.C2_nM arrays
        for kpt in self.wfs.kpt_u:
            kpt.C2_nM = np.empty_like(kpt.C_nM)

    def propagate(self, time, time_step):
        get_H_MM = self.hamiltonian.get_hamiltonian_matrix
        # --------------
        # Predictor step
        # --------------
        # 1. Store current C_nM
        self.save_wfs()  # kpt.C2_nM = kpt.C_nM
        for kpt in self.wfs.kpt_u:
            # H_MM(t) = <M|H(t)|M>
            kpt.H0_MM = get_H_MM(kpt, time)
            # 2. Solve Psi(t+dt) from
            #    (S_MM - 0.5j*H_MM(t)*dt) Psi(t+dt)
            #       = (S_MM + 0.5j*H_MM(t)*dt) Psi(t)
            self.propagate_wfs(kpt.C_nM, kpt.C_nM, kpt.S_MM, kpt.H0_MM,
                               time_step)
        # ---------------
        # Propagator step
        # ---------------
        # 1. Calculate H(t+dt)
        self.hamiltonian.update()
        for kpt in self.wfs.kpt_u:
            # 2. Estimate H(t+0.5*dt) ~ 0.5 * [ H(t) + H(t+dt) ]
            kpt.H0_MM += get_H_MM(kpt, time + time_step)
            kpt.H0_MM *= 0.5
            # 3. Solve Psi(t+dt) from
            #    (S_MM - 0.5j*H_MM(t+0.5*dt)*dt) Psi(t+dt)
            #       = (S_MM + 0.5j*H_MM(t+0.5*dt)*dt) Psi(t)
            self.propagate_wfs(kpt.C2_nM, kpt.C_nM, kpt.S_MM, kpt.H0_MM,
                               time_step)
            kpt.H0_MM = None
        # 4. Calculate new Hamiltonian (and density)
        self.hamiltonian.update()
        return time + time_step

    def save_wfs(self):
        for kpt in self.wfs.kpt_u:
            kpt.C2_nM[:] = kpt.C_nM

    def todict(self):
        return {'name': 'sicn'}


class TaylorPropagator(Propagator):

    def __init__(self):
        Propagator.__init__(self)
        raise NotImplementedError('TaylorPropagator not implemented')

    def initialize(self, paw):
        if 1:
            # XXX to propagator class
            if self.propagator == 'taylor' and self.blacs:
                # cholS_mm = self.mm_block_descriptor.empty(dtype=complex)
                for kpt in self.wfs.kpt_u:
                    kpt.invS_MM = kpt.S_MM.copy()
                    scalapack_inverse(self.mm_block_descriptor,
                                      kpt.invS_MM, 'L')
            if self.propagator == 'taylor' and not self.blacs:
                tmp = inv(self.wfs.kpt_u[0].S_MM)
                self.wfs.kpt_u[0].invS = tmp

    def taylor_propagator(self, sourceC_nM, targetC_nM, S_MM, H_MM, dt):
        self.timer.start('Taylor propagator')

        if self.blacs:
            # XXX, Preallocate
            target_blockC_nm = self.Cnm_block_descriptor.empty(dtype=complex)
            if self.density.gd.comm.rank != 0:
                # XXX Fake blacks nbands, nao, nbands, nao grid because some
                # weird asserts
                # (these are 0,x or x,0 arrays)
                sourceC_nM = self.CnM_unique_descriptor.zeros(dtype=complex)

            # Zeroth order taylor to target
            self.CnM2nm.redistribute(sourceC_nM, target_blockC_nm)

            # XXX, preallocate, optimize use of temporal arrays
            temp_blockC_nm = target_blockC_nm.copy()
            temp2_blockC_nm = target_blockC_nm.copy()

            order = 4
            assert self.wfs.kd.comm.size == 1
            for n in range(order):
                # Multiply with hamiltonian
                pblas_simple_hemm(self.mm_block_descriptor,
                                  self.Cnm_block_descriptor,
                                  self.Cnm_block_descriptor,
                                  H_MM,
                                  temp_blockC_nm,
                                  temp2_blockC_nm, side='R')
                # XXX: replace with not simple gemm
                temp2_blockC_nm *= -1j * dt / (n + 1)
                # Multiply with inverse overlap
                pblas_simple_hemm(self.mm_block_descriptor,
                                  self.Cnm_block_descriptor,
                                  self.Cnm_block_descriptor,
                                  self.wfs.kpt_u[0].invS_MM,  # XXX
                                  temp2_blockC_nm,
                                  temp_blockC_nm, side='R')
                target_blockC_nm += temp_blockC_nm
            if self.density.gd.comm.rank != 0:  # Todo: Change to gd.rank
                # XXX Fake blacks nbands, nao, nbands, nao grid because
                # some weird asserts
                # (these are 0,x or x,0 arrays)
                target = self.CnM_unique_descriptor.zeros(dtype=complex)
            else:
                target = targetC_nM
            self.Cnm2nM.redistribute(target_blockC_nm, target)

            self.density.gd.comm.broadcast(targetC_nM, 0)
        else:
            assert self.wfs.kd.comm.size == 1
            if self.density.gd.comm.rank == 0:
                targetC_nM[:] = sourceC_nM[:]
                tempC_nM = sourceC_nM.copy()
                order = 4
                for n in range(order):
                    tempC_nM[:] = \
                        np.dot(self.wfs.kpt_u[0].invS,
                               np.dot(H_MM, 1j * dt / (n + 1) *
                                      tempC_nM.T.conjugate())).T.conjugate()
                    targetC_nM += tempC_nM
            self.density.gd.comm.broadcast(targetC_nM, 0)

        self.timer.stop('Taylor propagator')



from __future__ import annotations

from typing import Dict, Tuple

import numpy as np
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.new import zip
from gpaw.new.lcao.wave_functions import LCAOWaveFunctions
from gpaw.new.potential import Potential
from gpaw.typing import Array2D, Array3D
from gpaw.utilities.blas import mmm

Derivatives = Tuple[Array3D,
                    Array3D,
                    Dict[int, Array3D]]


class TCIDerivatives:
    def __init__(self, manytci, atomdist, nao: int):
        self.manytci = manytci
        self.atomdist = atomdist
        self.nao = nao

        self._derivatives_q: dict[int, Derivatives] = {}

    def calculate_derivatives(self,
                              q: int) -> Derivatives:
        if not self._derivatives_q:
            dThetadR_qvMM, dTdR_qvMM = self.manytci.O_qMM_T_qMM(
                self.atomdist.comm,
                0, self.nao,
                False, derivative=True)

            self.atomdist.comm.sum(dThetadR_qvMM)
            self.atomdist.comm.sum(dTdR_qvMM)

            dPdR_aqvMi = self.manytci.P_aqMi(
                self.atomdist.indices,
                derivative=True)

            dPdR_qavMi = [{a: dPdR_qvMi[q]
                           for a, dPdR_qvMi in dPdR_aqvMi.items()}
                          for q in range(len(dThetadR_qvMM))]

            self._derivatives_q = {
                q: (dThetadR_vMM, dTdR_vMM, dPdR_avMi)
                for q, (dThetadR_vMM, dTdR_vMM, dPdR_avMi)
                in enumerate(zip(dThetadR_qvMM, dTdR_qvMM, dPdR_qavMi))}

        return self._derivatives_q[q]


def add_force_contributions(wfs: LCAOWaveFunctions,
                            potential: Potential,
                            F_av: Array2D) -> None:
    (dThetadR_vMM,
     dTdR_vMM,
     dPdR_avMi) = wfs.tci_derivatives.calculate_derivatives(wfs.q)

    indices = []
    M1 = 0
    for a, sphere in enumerate(wfs.basis.sphere_a):
        M2 = M1 + sphere.Mmax
        indices.append((a, M1, M2))
        M1 = M2

    setups = wfs.setups

    rhoT_MM = wfs.calculate_density_matrix().T
    erhoT_MM = wfs.calculate_density_matrix(eigs=True).T

    add_kinetic_term(rhoT_MM, dTdR_vMM, F_av, indices)
    add_pot_term(potential.vt_sR[wfs.spin], wfs.basis, wfs.q, rhoT_MM, F_av)
    add_den_mat_term(erhoT_MM, dThetadR_vMM, F_av, indices)
    for b in wfs.atomdist.indices:
        add_den_mat_paw_term(b,
                             setups[b].dO_ii,
                             wfs.P_aMi[b],
                             dPdR_avMi[b],
                             erhoT_MM,
                             indices,
                             F_av)
        add_atomic_density_term(b,
                                potential.dH_asii[b][wfs.spin],
                                wfs.P_aMi[b],
                                dPdR_avMi[b],
                                rhoT_MM,
                                indices,
                                F_av)


def add_kinetic_term(rhoT_MM, dTdR_vMM, F_av, indices):
    """Calculate Kinetic energy term in LCAO

    :::

                      dT
     _a        --- --   μν
     F += 2 Re >   >  ---- ρ
               --- --  _    νμ
               μ=a ν  dR
                        μν
            """

    for a, M1, M2 in indices:
        F_av[a, :] += 2 * np.einsum('vmM, Mm -> v',
                                    dTdR_vMM[:, M1:M2],
                                    rhoT_MM[:, M1:M2]).real


def add_pot_term(vt_R: UniformGridFunctions,
                 basis,
                 q: int,
                 rhoT_MM,
                 F_av) -> None:
    """Calculate potential term"""
    # Potential contribution
    #
    #           -----      /  d Phi  (r)
    #  a         \        |        mu    ~
    # F += -2 Re  )       |   ---------- v (r)  Phi  (r) dr rho
    #            /        |     d R                nu          nu mu
    #           -----    /         a
    #        mu in a; nu
    #
    F_av += basis.calculate_force_contribution(vt_R.data,
                                               rhoT_MM,
                                               q)


def add_den_mat_term(erhoT_MM, dThetadR_vMM, F_av, indices):
    """Calculate density matrix term in LCAO"""
    # Density matrix contribution due to basis overlap
    #
    #            ----- d Theta
    #  a          \           mu nu
    # F  += -2 Re  )   ------------  E
    #             /        d R        nu mu
    #            -----        mu nu
    #         mu in a; nu
    #
    for a, M1, M2 in indices:
        F_av[a, :] -= 2 * np.einsum('vmM, Mm -> v',
                                    dThetadR_vMM[:, M1:M2],
                                    erhoT_MM[:, M1:M2]).real


def add_den_mat_paw_term(b, dO_ii, P_Mi, dPdR_vMi, erhoT_MM, indices, F_av):
    """Calcualte PAW correction"""
    # TO DO: split this function into
    # _get_den_mat_paw_term (which calculate Frho_av) and
    # get_paw_correction (which calculate ZE_MM)
    # Density matrix contribution from PAW correction
    #
    #           -----                        -----
    #  a         \      a                     \     b
    # F +=  2 Re  )    Z      E        - 2 Re  )   Z      E
    #            /      mu nu  nu mu          /     mu nu  nu mu
    #           -----                        -----
    #           mu nu                    b; mu in a; nu
    #
    # with
    #                  b*
    #         -----  dP
    #   b      \       i mu    b   b
    #  Z     =  )   -------- dS   P
    #   mu nu  /     dR        ij  j nu
    #         -----    b mu
    #           ij
    #
    dtype = P_Mi.dtype
    Z_MM = np.zeros((len(P_Mi), len(P_Mi)), dtype)
    dOP_iM = np.zeros((len(dO_ii), len(P_Mi)), dtype)
    mmm(1.0, dO_ii.astype(dtype), 'N', P_Mi, 'C', 0.0, dOP_iM)
    for v in range(3):
        mmm(1.0,
            dPdR_vMi[v], 'N',
            dOP_iM, 'N',
            0.0, Z_MM)
        ZE_M = np.einsum('MN, MN -> M', Z_MM, erhoT_MM).real
        for a, M1, M2 in indices:
            dE = 2 * ZE_M[M1:M2].sum()
            F_av[a, v] -= dE  # the "b; mu in a; nu" term
            F_av[b, v] += dE  # the "mu nu" term


def add_atomic_density_term(b, dH_ii, P_Mi, dPdR_vMi, rhoT_MM, indices, F_av):
    # Atomic density contribution
    #            -----                         -----
    #  a          \     a                       \     b
    # F  += -2 Re  )   A      rho       + 2 Re   )   A      rho
    #             /     mu nu    nu mu          /     mu nu    nu mu
    #            -----                         -----
    #            mu nu                     b; mu in a; nu
    #
    #                  b*
    #         ----- d P
    #  b       \       i mu   b   b
    # A     =   )   ------- dH   P
    #  mu nu   /    d R       ij  j nu
    #         -----    b mu
    #           ij
    #
    dtype = P_Mi.dtype
    A_MM = np.zeros((len(P_Mi), len(P_Mi)), dtype)
    dHP_iM = np.zeros((len(dH_ii), len(P_Mi)), dtype)
    mmm(1.0, dH_ii.astype(dtype), 'N', P_Mi, 'C', 0.0, dHP_iM)
    for v in range(3):
        mmm(1.0,
            dPdR_vMi[v], 'N',
            dHP_iM, 'N',
            0.0, A_MM)
        AR_M = np.einsum('MN, MN -> M', A_MM, rhoT_MM).real
        for a, M1, M2 in indices:
            dE = 2 * AR_M[M1:M2].sum()
            F_av[a, v] += dE  # the "b; mu in a; nu" term
            F_av[b, v] -= dE  # the "mu nu" term



import os

from ase.units import Hartree

from gpaw.lrtddft.spectrum import spectrum, rotatory_spectrum


def check_convergence(lr,             # LrTDDFT object
                      dirname='conv',  # directory name to store the files
                      title=None,     # title for gnuplot
                      dn=None,        # steps to vary istart/jend
                      dE=None,        # steps to vary energy range
                      emin=None,
                      emax=None,
                      folding='Gauss',
                      istart=None,
                      jend=None,
                      width=0.03):     # Gauss/Lorentz width
    """Study the convergence of a LrTDDFT calcualtion by varying istart/jend.
    A gnuplot file will be created with the name 'dirname'/conv.gpl."""

    if istart is None:
        istart0 = lr.kss.restrict['istart']
    else:
        if istart < lr.kss.restrict['istart']:
            raise RuntimeError
        istart0 = istart
    if jend is None:
        jend0 = lr.kss.restrict['jend']
    else:
        if jend > lr.kss.restrict['jend']:
            raise RuntimeError
        jend0 = jend

    # create subdirectory for the files
    if not os.path.isdir(dirname):
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        else:
            raise RuntimeError('Can\'t create directory ' + dirname)

    def fname(filename):
        return dirname + '/' + filename

    with open(fname('conv.gpl'), 'w') as fgpl:
        print('set xlabel "omega [eV]"', file=fgpl)
        print('set ylabel "Folded osc. strength [1/eV]"', file=fgpl)
        if not emin:
            emin_gpl = '*'
        else:
            emin_gpl = str(emin)
        if not emax:
            emax_gpl = '*'
        else:
            emax_gpl = str(emax)
        print('set xrange [' + emin_gpl + ':' + emax_gpl + ']', file=fgpl)
        if title:
            print('set title "' + str(title) + '"', file=fgpl)

        # kss
        spectrum(lr.kss, fname('kss.dat'), width=width)
        spectrum(lr.kss, fname('ksssticks.dat'), folding=None)

        # full
        lr.diagonalize(restrict=dict(istart=istart0, jend=jend0))
        spectrum(lr, fname('full.dat'), width=width)
        spectrum(lr, fname('fullsticks.dat'), folding=None)
        print('plot "' + fname('full.dat') + '" t "full" w l lt 1, \\',
              file=fgpl)
        print('     "' + fname('fullsticks.dat') +
              '" u 1:($2*20) t "" w impulses lt 1, \\', file=fgpl)
        print('     "' + fname('kss.dat') + '" t "Kohn-Sham" w l lt 2',
              file=fgpl)
        print('pause -10', file=fgpl)

        if dn is None:
            dn = -istart0 + jend0
            dn = int(dn / 10.)

        if dE is None:
            # vary istart
            print('plot "' + fname('full.dat') + '" t "istart=' +
                  str(istart0) + '" w l lt 1, \\', file=fgpl)
            for i in range(1, 4):
                istart = istart0 + i * dn
                lr.diagonalize(restrict=dict(istart=istart, jend=jend0))
                fn = fname('istart' + str(istart) + '.dat')
                spectrum(lr, fn, width=width)
                print('    "' + fn + '" t "istart=' + str(istart) +
                      '" w l lt', i + 1, end=' ', file=fgpl)
                if i < 3:
                    print(', \\', end=' ', file=fgpl)
                print(file=fgpl)
            print('pause -10', file=fgpl)

            # vary jend
            print('plot "' + fname('full.dat') + '" t "jend=' +
                  str(jend0) + '" w l lt 1, \\', file=fgpl)
            for i in range(1, 4):
                jend = jend0 - i * dn
                lr.diagonalize(restrict=dict(jend=jend, istart=istart0))
                fn = fname('jend' + str(jend) + '.dat')
                spectrum(lr, fn, width=width)
                print('    "' + fn + '" t "jend=' + str(jend) +
                      '" w l lt', i + 1, end=' ', file=fgpl)
                if i < 3:
                    print(', \\', end=' ', file=fgpl)
                print(file=fgpl)
            print('pause -10', file=fgpl)
        else:
            # vary the enrgy range
            max_kss_energy = 0.
            for kss in lr.Om.kss:
                max_kss_energy = max(max_kss_energy,
                                     kss.get_energy() * Hartree)
            print('plot "' + fname('full.dat') + '" t "full"' +
                  ' w l lt 1, \\', file=fgpl)
            for i in range(1, 4):
                max_energy = max_kss_energy - i * dE
                lr.diagonalize(restrict=dict(energy_range=max_energy))
                fn = fname('max_energy' + str(max_energy) + '.dat')
                spectrum(lr, fn, width=width)
                print('    "' + fn + '" t "dE=-' + str(
                    i * dE) + '" w l lt', end=' ', file=fgpl)
                print(i + 1, end=' ', file=fgpl)
                if i < 3:
                    print(', \\', end=' ', file=fgpl)
                print(file=fgpl)
            print('pause -10', file=fgpl)

        # plot different directions
        print('plot "' + fname('full.dat') +
              '" u 1:3 t "x" w l lt 1, \\', file=fgpl)
        print('     "' + fname('full.dat') +
              '" u 1:4 t "y" w l lt 2, \\', file=fgpl)
        print('     "' + fname('full.dat') + '" u 1:5 t "z" w l lt 3',
              file=fgpl)
        print('pause -10', file=fgpl)

        # plot rotary strength
        if lr[0].magn is not None:
            print('set ylabel "Folded rot. strength [cgs/eV]"', file=fgpl)
            rotatory_spectrum(lr, fname('rotatory.dat'), width=width)
            rotatory_spectrum(lr, fname('rotatory_sticks.dat'), folding=None)
            print('plot "' + fname('rotatory.dat') +
                  '" t "rotatory" w l lt 1, \\', file=fgpl)
            print('     "' + fname('rotatory_sticks.dat') +
                  '" u 1:($2*20) t "" w impulses lt 1', file=fgpl)
            print('pause -10', file=fgpl)



from math import pi

import numpy as np
import pytest
from gpaw.spherical_harmonics import (YL, Y, Yl, gam, print_YL_table_code,
                                      write_c_code)


def yLL(L1, L2):
    s = 0.0
    for c1, n1 in YL[L1]:
        for c2, n2 in YL[L2]:
            s += c1 * c2 * gam(n1[0] + n2[0], n1[1] + n2[1], n1[2] + n2[2])
    return s


def test_yy():
    Lmax = len(YL)
    for L1 in range(Lmax):
        for L2 in range(Lmax):
            r = 0.0
            if L1 == L2:
                r = 1.0
            assert yLL(L1, L2) == pytest.approx(r, abs=1e-14)


def test_y_c_code():
    R = np.zeros(3)
    Y = np.zeros(1)
    Yl(0, R, Y)
    assert Y[0] == pytest.approx((4 * pi)**-0.5)
    Y = np.zeros(2 * 8 + 1)
    with pytest.raises(RuntimeError):
        Yl(8, R, Y)


def test_y_c_code2():
    R_c = np.array([0.1, -0.2, 0.3])
    for l in range(8):
        print('l', l)

        # Using C implementation
        rlY1_m = np.zeros(2 * l + 1) + np.nan
        Yl(l, R_c, rlY1_m)
        print(rlY1_m)

        # Using Python implementation
        rlY2_m = []
        for m in range(-l, l + 1):
            L = l**2 + l + m
            rlY2_m.append(Y(L, *R_c))
        rlY2_m = np.array(rlY2_m)
        print(rlY2_m)

        assert np.allclose(rlY2_m, rlY1_m)


def test_write_c_code(capsys):
    write_c_code(1)
    s = capsys.readouterr().out
    s = s.replace(' ', '')
    s = s.replace('\n', '')
    assert s == (
        'elseif(l==1)' +
        '{Y_m[0]=0.4886025119029199*y;' +
        'Y_m[1]=0.4886025119029199*z;' +
        'Y_m[2]=0.4886025119029199*x;}')


def test_print_YL_table_code():
    pytest.importorskip('sympy')
    print_YL_table_code()



import pickle
from math import log, pi

import numpy as np
from gpaw.overlap import Overlap
from ase.units import Hartree
from gpaw.utilities.cg import CG
import gpaw.mpi as mpi


class XAS:
    def __init__(self, paw, mode='xas', center=None, spin=0):
        wfs = paw.wfs
        self.orthogonal = wfs.gd.orthogonal
        self.cell_cv = np.array(wfs.gd.cell_cv)
        assert wfs.world.size == 1  # assert not mpi.parallel
        # assert wfs.gd.orthogonal

        # to allow spin polarized calclulation
        nkpts = len(wfs.kd.ibzk_kc)

        # the following lines are to stop the user to make mistakes
        # if mode is not 'xes' and spin == 1:
        #     raise RuntimeError(
        #         'The core hole is always in spin 0: please use spin=0')

        if wfs.nspins == 1:
            if spin != 0:
                raise RuntimeError(
                    'use spin=0 for a spin paired calculation')
            nocc = wfs.setups.nvalence // 2
            self.list_kpts = range(nkpts)
        else:
            self.list_kpts = []

            if spin != 0 and spin != 1:
                print('spin', spin)
                raise RuntimeError(
                    'use either spin=0 or spin=1')

            # find kpoints with correct spin
            for i, kpt in enumerate(wfs.kpt_u):
                if kpt.s == spin:
                    self.list_kpts.append(i)
                print(self.list_kpts)
            assert len(self.list_kpts) == nkpts

            # find number of occupied orbitals, if no fermi smearing
            nocc = 0.0
            for i in self.list_kpts:
                nocc += sum(wfs.kpt_u[i].f_n)
            nocc = int(nocc + 0.5)
            print('nocc', nocc)

        # look for the center with the corehole
        if center is not None:
            setup = wfs.setups[center]
            a = center
        else:
            for a, setup in enumerate(wfs.setups):
                if setup.phicorehole_g is not None:
                    break

        A_ci = setup.A_ci

        # xas, xes or all modes
        if mode == 'xas':
            n_start = nocc
            n_end = wfs.bd.nbands
            n = wfs.bd.nbands - nocc
        elif mode == 'xes':
            n_start = 0
            n_end = nocc
            n = nocc
        elif mode == 'all':
            n_start = 0
            n_end = wfs.bd.nbands
            n = wfs.bd.nbands
        else:
            raise RuntimeError(
                "wrong keyword for 'mode', use 'xas', 'xes' or 'all'")

        self.n = n

        self.eps_n = np.empty(nkpts * n)
        self.sigma_cn = np.empty((3, nkpts * n), complex)
        n1 = 0
        for kpt in wfs.kpt_u:
            if kpt.s != spin:
                continue

            n2 = n1 + n
            self.eps_n[n1:n2] = kpt.eps_n[n_start:n_end] * Hartree
            P_ni = kpt.P_ani[a][n_start:n_end]
            a_cn = np.inner(A_ci, P_ni)
            weight = kpt.weight * wfs.nspins / 2
            print('weight', weight)
            print(a_cn.shape, self.sigma_cn.shape)
            self.sigma_cn[:, n1:n2] = weight**0.5 * a_cn  # .real
            n1 = n2

        self.symmetry = wfs.kd.symmetry

    def get_spectra(self, fwhm=0.5, E_in=None, linbroad=None,
                    N=1000, kpoint=None,
                    proj=None, proj_xyz=True, stick=False):
        """Calculate spectra.

        Parameters:

        fwhm:
          the full width half maximum in eV for gaussian broadening
        linbroad:
          a list of three numbers, the first fwhm2, the second the value
          where the linear increase starts and the third the value where
          the broadening has reached fwhm2. example [0.5, 540, 550]
        E_in:
          a list of energy values where the spectrum is to be computed
          if None the orbital energies will be used to compute the energy
          range
        N:
          the number of bins in the broadened spectrum. If E_in is given N
          has no effect
        kpoint:
          select a specific k-point to calculate spectrum for
        proj:
          a list of vectors to project the transition dipole on. Default
          is None then only x,y,z components are calculated.  a_stick and
          a_c squares of the transition moments in resp. direction
        proj_xyz:
          if True keep projections in x, y and z. a_stck and a_c will have
          length 3 + len(proj). if False only those projections
          defined by proj keyword, a_stick and a_c will have length len(proj)
        stick:
          if False return broadened spectrum, if True return stick spectrum

        Symmtrization has been moved inside get_spectra because we want to
        symmtrice squares of transition dipoles."""

        # eps_n = self.eps_n[k_in*self.n: (k_in+1)*self.n -1]

        # proj keyword, check normalization of incoming vectors
        if proj_xyz:
            proj_3 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], float)
        else:
            proj_3 = np.array([], float)

        if proj is not None:
            assert self.orthogonal
            proj_2 = np.array(proj, float)
            if len(proj_2.shape) == 1:
                proj_2 = np.array([proj], float)

            for i, p in enumerate(proj_2):
                if sum(p ** 2) ** 0.5 != 1.0:
                    print('proj_2 %s not normalized' % i)
                    proj_2[i] /= sum(p ** 2) ** 0.5

            proj_tmp = np.zeros((proj_3.shape[0] + proj_2.shape[0], 3), float)

            for i, p in enumerate(proj_3):
                proj_tmp[i, :] = proj_3[i, :]

            for i, p in enumerate(proj_2):
                proj_tmp[proj_3.shape[0] + i, :] = proj_2[i, :]

            proj_3 = proj_tmp.copy()

        # now symmetrize
        sigma2_cn = np.zeros((proj_3.shape[0], self.sigma_cn.shape[1]), float)

        if self.symmetry is not None:
            for i, p in enumerate(proj_3):
                for op_cc in self.symmetry.op_scc:
                    op_vv = np.dot(np.linalg.inv(self.cell_cv),
                                   np.dot(op_cc, self.cell_cv))
                    s_tmp = np.dot(p, np.dot(op_vv, self.sigma_cn))
                    sigma2_cn[i, :] += (s_tmp * np.conjugate(s_tmp)).real
            sigma2_cn /= len(self.symmetry.op_scc)

        else:
            for i, p in enumerate(proj_3):
                s_tmp = np.dot(p, self.sigma_cn)
                sigma2_cn[i, :] += (s_tmp * np.conjugate(s_tmp)).real

        eps_n = self.eps_n[:]

        if kpoint is not None:
            eps_start = kpoint * self.n
            eps_end = (kpoint + 1) * self.n
        else:
            eps_start = 0
            eps_end = len(self.eps_n)

        # return stick spectrum if stick=True
        if stick:
            e_stick = eps_n[eps_start:eps_end]
            a_stick = sigma2_cn[:, eps_start:eps_end]

            return e_stick, a_stick

        # else return broadened spectrum
        else:
            if E_in is not None:
                e = np.array(E_in)
            else:
                emin = min(eps_n) - 2 * fwhm
                emax = max(eps_n) + 2 * fwhm
                e = emin + np.arange(N + 1) * ((emax - emin) / N)

            a_c = np.zeros((len(sigma2_cn), len(e)))

            if linbroad is None:
                # constant broadening fwhm
                alpha = 4 * log(2) / fwhm**2

                for n, eps in enumerate(eps_n[eps_start:eps_end]):
                    x = -alpha * (e - eps)**2
                    x = np.clip(x, -100.0, 100.0)
                    a_c += np.outer(sigma2_cn[:, n + eps_start],
                                    (alpha / pi)**0.5 * np.exp(x))
            else:

                # constant broadening fwhm until linbroad[1] and a
                # constant broadening over linbroad[2] with fwhm2=
                # linbroad[0]
                fwhm2 = linbroad[0]
                lin_e1 = linbroad[1]
                lin_e2 = linbroad[2]
                print('fwhm', fwhm, fwhm2, lin_e1, lin_e2)
                for n, eps in enumerate(eps_n):
                    if eps < lin_e1:
                        alpha = 4 * log(2) / fwhm**2
                    elif eps <= lin_e2:
                        fwhm_lin = (fwhm + (eps - lin_e1) *
                                    (fwhm2 - fwhm) / (lin_e2 - lin_e1))
                        alpha = 4 * log(2) / fwhm_lin**2
                    elif eps >= lin_e2:
                        alpha = 4 * log(2) / fwhm2**2

                    x = -alpha * (e - eps)**2
                    x = np.clip(x, -100.0, 100.0)
                    a_c += np.outer(sigma2_cn[:, n],
                                    (alpha / pi)**0.5 * np.exp(x))

            return e, a_c


class RecursionMethod:
    """This class implements the Haydock recursion method. """

    def __init__(self, paw=None, filename=None,
                 tol=1e-10, maxiter=100, proj=None,
                 proj_xyz=True):

        if paw is not None:
            wfs = paw.wfs
            assert wfs.gd.orthogonal

            self.wfs = wfs
            self.hamiltonian = paw.hamiltonian
            self.nkpts = len(wfs.kd.ibzk_kc) * wfs.nspins
            self.nmykpts = len(wfs.kpt_u)

            self.k1 = wfs.kd.comm.rank * self.nmykpts
            self.k2 = self.k1 + self.nmykpts

            print('k1', self.k1, 'k2', self.k2)

            # put spin and weight index in the columns corresponding
            # to this processors k-points
            self.spin_k = np.zeros(self.nkpts, int)
            self.weight_k = np.zeros(self.nkpts)

            for n, i in enumerate(range(self.k1, self.k2)):
                self.spin_k[i] = wfs.kpt_u[n].s
                self.weight_k[i] = wfs.kpt_u[n].weight

            self.op_scc = None
            if wfs.kd.symmetry is not None:
                self.op_scc = wfs.kd.symmetry.op_scc
        else:
            self.k1 = 0
            self.k2 = None
            self.wfs = None
            wfs = None

        self.tol = tol
        self.maxiter = maxiter

        if filename is not None:
            self.read(filename)
            if wfs is not None:
                self.allocate_tmp_arrays()
        else:
            self.initialize_start_vector(proj=proj, proj_xyz=proj_xyz)

    def read(self, filename):
        with open(filename, 'rb') as fd:
            data = pickle.load(fd)
        self.nkpts = data['nkpts']
        if 'swaps' in data:
            # This is an old file:
            self.op_scc = np.array([np.identity(3, int)[list(swap)]
                                    for swap in data['swaps']])
        else:
            self.op_scc = data['symmetry operations']
        self.weight_k = data['weight_k']
        self.spin_k = data['spin_k']
        self.dim = data['dim']
        k1, k2 = self.k1, self.k2
        if k2 is None:
            k2 = self.nkpts
        a_kci, b_kci = data['ab']
        self.a_uci = a_kci[k1:k2].copy()
        self.b_uci = b_kci[k1:k2].copy()

        if self.wfs is not None and 'arrays' in data:
            print('reading arrays')
            w_kcG, wold_kcG, y_kcG = data['arrays']
            i = [slice(k1, k2), slice(0, self.dim)] + self.wfs.gd.get_slice()
            self.w_ucG = w_kcG[i].copy()
            self.wold_ucG = wold_kcG[i].copy()
            self.y_ucG = y_kcG[i].copy()

    def write(self, filename, mode=''):
        assert self.wfs is not None
        kpt_comm = self.wfs.kd.comm
        gd = self.wfs.gd

        if gd.comm.rank == 0:
            if kpt_comm.rank == 0:
                nmyu, dim, ni = self.a_uci.shape
                a_kci = np.empty((kpt_comm.size, nmyu, dim, ni),
                                 self.wfs.dtype)
                b_kci = np.empty((kpt_comm.size, nmyu, dim, ni),
                                 self.wfs.dtype)

                kpt_comm.gather(self.a_uci, 0, a_kci)
                kpt_comm.gather(self.b_uci, 0, b_kci)
                kpt_comm.sum(self.spin_k, 0)
                kpt_comm.sum(self.weight_k, 0)

                a_kci.shape = (self.nkpts, dim, ni)
                b_kci.shape = (self.nkpts, dim, ni)
                data = {'ab': (a_kci, b_kci),
                        'nkpts': self.nkpts,
                        'symmetry operations': self.op_scc,
                        'weight_k': self.weight_k,
                        'spin_k': self.spin_k,
                        'dim': dim}
            else:
                kpt_comm.gather(self.a_uci, 0)
                kpt_comm.gather(self.b_uci, 0)
                kpt_comm.sum(self.spin_k, 0)
                kpt_comm.sum(self.weight_k, 0)

        if mode == 'all':
            w0_ucG = gd.collect(self.w_ucG)
            wold0_ucG = gd.collect(self.wold_ucG)
            y0_ucG = gd.collect(self.y_ucG)
            if gd.comm.rank == 0:
                if kpt_comm.rank == 0:
                    w_kcG = gd.empty((self.nkpts, dim), self.wfs.dtype,
                                     global_array=True)
                    wold_kcG = gd.empty((self.nkpts, dim), self.wfs.dtype,
                                        global_array=True)
                    y_kcG = gd.empty((self.nkpts, dim), self.wfs.dtype,
                                     global_array=True)
                    kpt_comm.gather(w0_ucG, 0, w_kcG)
                    kpt_comm.gather(wold0_ucG, 0, wold_kcG)
                    kpt_comm.gather(y0_ucG, 0, y_kcG)
                    data['arrays'] = (w_kcG, wold_kcG, y_kcG)
                else:
                    kpt_comm.gather(w0_ucG, 0)
                    kpt_comm.gather(wold0_ucG, 0)
                    kpt_comm.gather(y0_ucG, 0)

        if self.wfs.world.rank == 0:
            with open(filename, 'wb') as fd:
                pickle.dump(data, fd)

    def allocate_tmp_arrays(self):

        self.tmp1_cG = self.wfs.gd.zeros(self.dim, self.wfs.dtype)
        self.tmp2_cG = self.wfs.gd.zeros(self.dim, self.wfs.dtype)
        self.z_cG = self.wfs.gd.zeros(self.dim, self.wfs.dtype)

    def initialize_start_vector(self, proj=None, proj_xyz=True):
        # proj is one list of vectors [[e1_x,e1_y,e1_z],[e2_x,e2_y,e2_z]]
        # ( or [ex,ey,ez] if only one projection )
        # that the spectrum will be projected on
        # default is to only calculate the averaged spectrum
        # if proj_xyz is True, keep projection in x,y,z, if False
        # only calculate the projections in proj

        # Create initial wave function:
        nmykpts = self.nmykpts

        for a, setup in enumerate(self.wfs.setups):
            if setup.phicorehole_g is not None:
                break
        A_ci = setup.A_ci

        #
        # proj keyword
        #

        # check normalization of incoming vectors
        if proj is not None:
            proj_2 = np.array(proj, float)
            if len(proj_2.shape) == 1:
                proj_2 = np.array([proj], float)

            for i, p in enumerate(proj_2):
                if sum(p ** 2) ** 0.5 != 1.0:
                    print('proj_2 %s not normalized' % i)
                    proj_2[i] /= sum(p ** 2) ** 0.5

            proj_tmp = []
            for p in proj_2:
                proj_tmp.append(np.dot(p, A_ci))
            proj_tmp = np.array(proj_tmp, float)

            # if proj_xyz is True, append projections to A_ci
            if proj_xyz:
                A_ci_tmp = np.zeros((3 + proj_2.shape[0], A_ci.shape[1]))
                A_ci_tmp[0:3, :] = A_ci
                A_ci_tmp[3:, :] = proj_tmp

            # otherwise, replace A_ci by projections
            else:
                A_ci_tmp = np.zeros((proj_2.shape[0], A_ci.shape[1]))
                A_ci_tmp = proj_tmp
            A_ci = A_ci_tmp

        self.dim = len(A_ci)

        self.allocate_tmp_arrays()

        self.w_ucG = self.wfs.gd.zeros((nmykpts, self.dim), self.wfs.dtype)
        self.wold_ucG = self.wfs.gd.zeros((nmykpts, self.dim), self.wfs.dtype)
        self.y_ucG = self.wfs.gd.zeros((nmykpts, self.dim), self.wfs.dtype)

        self.a_uci = np.zeros((nmykpts, self.dim, 0), self.wfs.dtype)
        self.b_uci = np.zeros((nmykpts, self.dim, 0), self.wfs.dtype)

        A_aci = self.wfs.pt.dict(3, zero=True)
        if a in A_aci:
            A_aci[a] = A_ci.astype(self.wfs.dtype)
        for u in range(nmykpts):
            self.wfs.pt.add(self.w_ucG[u], A_aci, u)

    def run(self, nsteps, inverse_overlap='exact'):

        if inverse_overlap == 'exact':
            self.solver = self.solve
        elif inverse_overlap == 'approximate':
            self.solver = self.solve2
        elif inverse_overlap == 'noinverse':
            self.solver = self.solve3
        else:
            raise RuntimeError("Error, inverse_solver must be either 'exact' "
                               "'approximate' or 'noinverse'")

        self.overlap = Overlap()

        ni = self.a_uci.shape[2]
        a_uci = np.empty((self.nmykpts, self.dim, ni + nsteps), self.wfs.dtype)
        b_uci = np.empty((self.nmykpts, self.dim, ni + nsteps), self.wfs.dtype)
        a_uci[:, :, :ni] = self.a_uci
        b_uci[:, :, :ni] = self.b_uci
        self.a_uci = a_uci
        self.b_uci = b_uci

        for u in range(self.nmykpts):
            for i in range(nsteps):
                self.step(u, ni + i)

    def step(self, u, i):
        print(u, i)
        integrate = self.wfs.gd.integrate
        w_cG = self.w_ucG[u]
        y_cG = self.y_ucG[u]
        wold_cG = self.wold_ucG[u]
        z_cG = self.z_cG

        self.solver(w_cG, self.z_cG, u)
        I_c = np.reshape(integrate(np.conjugate(z_cG) * w_cG)**-0.5,
                         (self.dim, 1, 1, 1))
        z_cG *= I_c
        w_cG *= I_c

        if i != 0:
            b_c = 1.0 / I_c
        else:
            b_c = np.reshape(np.zeros(self.dim), (self.dim, 1, 1, 1))

        self.hamiltonian.apply(z_cG, y_cG, self.wfs, self.wfs.kpt_u[u])
        a_c = np.reshape(integrate(np.conjugate(z_cG) * y_cG),
                         (self.dim, 1, 1, 1))
        wnew_cG = (y_cG - a_c * w_cG - b_c * wold_cG)
        wold_cG[:] = w_cG
        w_cG[:] = wnew_cG
        self.a_uci[u, :, i] = a_c[:, 0, 0, 0]
        self.b_uci[u, :, i] = b_c[:, 0, 0, 0]

    def continued_fraction(self, e, k, c, i, imax):
        a_i = self.a_uci[k, c]
        b_i = self.b_uci[k, c]
        if i == imax - 2:
            return self.terminator(a_i[i], b_i[i], e)
        return 1.0 / (a_i[i] - e -
                      b_i[i + 1]**2 *
                      self.continued_fraction(e, k, c, i + 1, imax))

    def get_spectra(self, eps_s, delta=0.1, imax=None, kpoint=None, fwhm=None,
                    linbroad=None, spin=0):
        assert not mpi.parallel

        # the following lines are to stop the user to make mistakes
        # if spin == 1:
        #     raise RuntimeError(
        #         'The core hole is always in spin 0: please use spin=0')

        n = len(eps_s)

        sigma_cn = np.zeros((self.dim, n))
        if imax is None:
            imax = self.a_uci.shape[2]
        eps_n = (eps_s + delta * 1.0j) / Hartree

        # if a certain k-point is chosen
        if kpoint is not None:
            for c in range(self.dim):
                sigma_cn[c] += self.continued_fraction(eps_n, kpoint, c,
                                                       0, imax).imag
        else:
            for k in range(self.nkpts):
                print('kpoint', k, 'spin_k', self.spin_k[k], spin,
                      'weight', self.weight_k[k])
                if self.spin_k[k] == spin:
                    weight = self.weight_k[k]
                    for c in range(self.dim):
                        sigma_cn[c] += weight * self.continued_fraction(
                            eps_n, k, c, 0, imax).imag

        if self.op_scc is not None:
            sigma0_cn = sigma_cn
            sigma_cn = np.zeros((self.dim, n))
            for op_cc in self.op_scc:
                sigma_cn += np.dot(op_cc**2, sigma0_cn)
            sigma_cn /= len(self.op_scc)

        # gaussian broadening
        if fwhm is not None:
            sigma_tmp = np.zeros(sigma_cn.shape)

            # constant broadening fwhm
            if linbroad is None:
                alpha = 4 * log(2) / fwhm**2
                for n, eps in enumerate(eps_s):
                    x = -alpha * (eps_s - eps)**2
                    x = np.clip(x, -100.0, 100.0)
                    sigma_tmp += np.outer(sigma_cn[:, n],
                                          (alpha / pi)**0.5 * np.exp(x))

            else:
                # constant broadening fwhm until linbroad[1] and a
                # constant broadening over linbroad[2] with fwhm2=
                # linbroad[0]
                fwhm2 = linbroad[0]
                lin_e1 = linbroad[1]
                lin_e2 = linbroad[2]
                for n, eps in enumerate(eps_s):
                    if eps < lin_e1:
                        alpha = 4 * log(2) / fwhm**2
                    elif eps <= lin_e2:
                        fwhm_lin = (fwhm + (eps - lin_e1) *
                                    (fwhm2 - fwhm) / (lin_e2 - lin_e1))
                        alpha = 4 * log(2) / fwhm_lin**2
                    elif eps >= lin_e2:
                        alpha = 4 * log(2) / fwhm2**2

                    x = -alpha * (eps_s - eps)**2
                    x = np.clip(x, -100.0, 100.0)
                    sigma_tmp += np.outer(sigma_cn[:, n],
                                          (alpha / pi)**0.5 * np.exp(x))
            sigma_cn = sigma_tmp

        return sigma_cn

    def solve(self, w_cG, z_cG, u):
        # exact inverse overlap
        self.overlap.apply_inverse(w_cG, self.tmp1_cG, self.wfs,
                                   self.wfs.kpt_u[u])
        self.u = u
        CG(self, z_cG, self.tmp1_cG,
           tolerance=self.tol, maxiter=self.maxiter)

    def solve2(self, w_cG, z_cG, u):
        # approximate inverse overlap
        self.overlap.apply_inverse(w_cG, z_cG, self.wfs, self.wfs.kpt_u[u])

        self.u = u

    def solve3(self, w_cG, z_cG, u):
        # no inverse overlap
        z_cG[:] = w_cG
        self.u = u

    def sum(self, a):
        self.wfs.gd.comm.sum(a)
        return a

    def __call__(self, in_cG, out_cG):
        """Function that is called by CG. It returns S~-1Sx_in in x_out
        """

        kpt = self.wfs.kpt_u[self.u]
        self.overlap.apply(in_cG, self.tmp2_cG, self.wfs, kpt)
        self.overlap.apply_inverse(self.tmp2_cG, out_cG, self.wfs, kpt)

    def terminator(self, a, b, e):
        """ Analytic formula to terminate the continued fraction from
        [R Haydock, V Heine, and M J Kelly,
        J Phys. C: Solid State Physics, Vol 8, (1975), 2591-2605]
        """

        return 0.5 * (e - a - ((e - a)**2 - 4 * b**2)**0.5 / b**2)

    def duplicate_coefficients(self, nsteps, ntimes):
        n1 = self.a_uci.shape[0]
        n2 = self.a_uci.shape[1]
        ni = self.a_uci.shape[2]
        type_code = self.a_uci.dtype.name  # typecode()
        a_uci = np.empty((n1, n2, ni + nsteps * ntimes), type_code)
        b_uci = np.empty((n1, n2, ni + nsteps * ntimes), type_code)
        a_uci[:, :, :ni] = self.a_uci
        b_uci[:, :, :ni] = self.b_uci

        ni1 = ni
        ni2 = ni + nsteps
        for i in range(ntimes):
            a_uci[:, :, ni1: ni2] = a_uci[:, :, ni - nsteps:ni]
            b_uci[:, :, ni1: ni2] = b_uci[:, :, ni - nsteps:ni]
            ni1 += nsteps
            ni2 += nsteps
        self.a_uci = a_uci
        self.b_uci = b_uci


def write_spectrum(a, b, filename):
    f = open(filename, 'w')
    print(f, a.shape, b.shape)

    for i in range(a.shape[0]):
        print('%g' % a[i], b[0, i] + b[1, i] + b[2, i], end=' ', file=f)
        for b2 in b:
            print('%g' % b2[i], end=' ', file=f)
        print(file=f)
    f.close()



import numpy as np
import pytest
from ase.build import bulk

from gpaw import GPAW, FermiDirac, Mixer


@pytest.mark.later
def test_pathological_lcao_spos_derivative(in_tmp_dir):
    atoms = bulk('Si', 'diamond', a=5.4834322363595565)
    atoms *= (3, 3, 3)
    atoms.calc = GPAW(
        gpts=(32, 32, 32),
        mixer=Mixer(0.5, 5, 50.0),
        txt='grumble.txt',
        kpts=(2, 1, 1),
        mode='lcao',
        basis='sz(dzp)',
        xc='oldLDA',
        occupations=FermiDirac(0.01))
    f = atoms.get_forces()
    fmax = np.abs(f).max()
    print('maxforce', fmax)
    assert fmax < 0.05  # 0.03 normally, 2.7 with bug
    # fmax can be converged much closer to 0 with better grid/SCF convergence.



from ase.build import bulk
from gpaw import GPAW, PW
import pytest


@pytest.mark.stress
def test_pseudopotential_ah(in_tmp_dir):
    si = bulk('Si', 'diamond', a=5.5, cubic=not True)
    si.calc = GPAW(mode=PW(200),
                   setups='ah',
                   kpts=(2, 2, 2))
    f = si.get_forces()
    assert f == pytest.approx(0.0)
    s = si.get_stress()
    assert s[3:] == pytest.approx(0.0)
    assert s[:3] == pytest.approx(0.299, abs=0.001)
    si.calc.write('Si.gpw', 'all')
    GPAW('Si.gpw')



"""Compare TPSS from scratch and from PBE"""
import pytest
from ase import Atoms
from gpaw import GPAW, MixerSum, Davidson


@pytest.mark.mgga
def test_mgga_mgga_sc():
    n = Atoms('N', magmoms=[3])
    n.center(vacuum=2.5)

    def getkwargs():
        return dict(eigensolver=Davidson(4), mixer=MixerSum(0.5, 5, 10.0))

    n.calc = GPAW(xc='TPSS', **getkwargs())
    e1 = n.get_potential_energy()

    n.calc = GPAW(xc='PBE', **getkwargs())
    n.get_potential_energy()
    n.calc = n.calc.new(xc='TPSS')
    e2 = n.get_potential_energy()
    err = abs(e2 - e1)
    print('Energy difference', err)
    assert err < 3e-5, err



import pytest
from gpaw.upf import UPFSetupData


def get(fname):
    s = UPFSetupData(fname)
    return s


@pytest.mark.skip(reason='TODO')
def test_psp_upf_h2o():
    s = get('O.pz-hgh.UPF')
    # upfplot(s.data, show=True)

    # dr = x[1] - x[0]

    from gpaw.atom.atompaw import AtomPAW

    if 0:
        # f = 1.0  # 1e-12
        c = AtomPAW('H', [[[2.0], [4.0]]],
                    # charge=1 - f,
                    h=0.04,
                    # setups='paw',
                    # setups='hgh',
                    setups={'H': s}
                    )

        import pylab as pl
        pl.plot(c.wfs.gd.r_g, c.hamiltonian.vt_sg[0])
        pl.show()
        raise SystemExit

    # print 'test v201 Au.pz-d-hgh.UPF'
    # s = UPFSetupData('Au.pz-d-hgh.UPF')
    # print 'v201 ok'

    # print 'test horrible version O.pz-mt.UPF'
    # s = UPFSetupData('O.pz-mt.UPF')
    # print 'horrible version ok, relatively speaking'

    if 1:
        from gpaw import GPAW, PoissonSolver
        from gpaw.utilities import h2gpts
        from ase.build import molecule

        # s = UPFSetupData('/home/askhl/parse-upf/h_lda_v1.uspp.F.UPF')

        upfsetups = {'H': UPFSetupData('H.pz-hgh.UPF'),
                     'O': UPFSetupData('O.pz-hgh.UPF')}

        system = molecule('H2O')
        system.center(vacuum=3.5)
        calc = GPAW(txt='-',
                    nbands=6,
                    setups=upfsetups,
                    # setups='paw',
                    # hund=True,
                    # mixer=MixerSum(0.1, 5, 20.0),
                    # eigensolver='cg',
                    # occupations=FermiDirac(0.1),
                    # charge=1-1e-12,
                    # eigensolver='rmm-diis',
                    gpts=h2gpts(0.12, system.get_cell(), idiv=8),
                    poissonsolver=PoissonSolver(relax='GS', eps=1e-7),
                    xc='oldLDA',
                    # nbands=4
                    )

        system.calc = calc
        system.get_potential_energy()



#!/usr/bin/env python
# Copyright (C) 2003-2020  CAMP
# Please see the accompanying LICENSE file for further information.

import os
import re
import sys
import warnings
from pathlib import Path
from subprocess import run
from sysconfig import get_platform

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext as _build_ext
from setuptools.command.develop import develop as _develop
from setuptools.command.install import install as _install

from config import build_interpreter, check_dependencies, write_configuration

assert sys.version_info >= (3, 7)


def warn_deprecated(msg):
    msg = f'\n\n{msg}\n\n'
    warnings.warn(msg, DeprecationWarning)


def raise_error(msg):
    msg = f'\n\n{msg}\n\n'
    raise ValueError(msg)


# Get the current version number:
txt = Path('gpaw/__init__.py').read_text()
version = re.search("__version__ = '(.*)'", txt)[1]
ase_version_required = re.search("__ase_version_required__ = '(.*)'", txt)[1]

description = 'GPAW: DFT and beyond within the projector-augmented wave method'
long_description = Path('README.rst').read_text()

# Deprecation check
for i, arg in enumerate(sys.argv):
    if arg.startswith('--customize='):
        custom = arg.split('=')[1]
        raise DeprecationWarning(
            f'Please set GPAW_CONFIG={custom} or place {custom} in ' +
            '~/.gpaw/siteconfig.py')

libraries = ['xc']
library_dirs = []
include_dirs = []
extra_link_args = []
extra_compile_args = ['-Wall', '-Wno-unknown-pragmas', '-std=c99']
runtime_library_dirs = []
extra_objects = []
define_macros = [('NPY_NO_DEPRECATED_API', '7'),
                 ('GPAW_NO_UNDERSCORE_CBLACS', None),
                 ('GPAW_NO_UNDERSCORE_CSCALAPACK', None),
                 ('GPAW_MPI_INPLACE', None)]
if os.getenv('GPAW_GPU'):
    define_macros.append(('GPAW_GPU_AWARE_MPI', '1'))
undef_macros = ['NDEBUG']

parallel_python_interpreter = False
compiler = None
mpi = None
noblas = False
nolibxc = False
fftw = False
scalapack = False
libvdwxc = False
elpa = False

# Advanced:
# If these are defined, they replace
# all the default args from setuptools
compiler_args = None
linker_so_args = None
linker_exe_args = None

# Search and store current git hash if possible
try:
    from ase.utils import search_current_git_hash
    githash = search_current_git_hash('gpaw')
    if githash is not None:
        define_macros += [('GPAW_GITHASH', githash)]
    else:
        print('.git directory not found. GPAW git hash not written.')
except ImportError:
    print('ASE not found. GPAW git hash not written.')

# User provided customizations:
gpaw_config = os.environ.get('GPAW_CONFIG')
if gpaw_config and not Path(gpaw_config).is_file():
    raise FileNotFoundError(gpaw_config)
for siteconfig in [gpaw_config,
                   'siteconfig.py',
                   '~/.gpaw/siteconfig.py']:
    if siteconfig is not None:
        path = Path(siteconfig).expanduser()
        if path.is_file():
            print('Reading configuration from', path)
            exec(path.read_text())
            break
else:  # no break
    if not noblas:
        libraries.append('blas')

# Deprecation check
if 'mpicompiler' in locals():
    mpicompiler = locals()['mpicompiler']
    msg = 'Please remove deprecated declaration of mpicompiler.'
    if mpicompiler is None:
        mpi = False
        msg += (' Define instead in siteconfig one of the following lines:'
                '\n\nmpi = False\nmpi = True')
    else:
        mpi = True
        compiler = mpicompiler
        msg += (' Define instead in siteconfig:'
                f'\n\nmpi = True\ncompiler = {repr(compiler)}')
    warn_deprecated(msg)


# If `mpi` was not set in siteconfig,
# it is enabled by default if `mpicc` is found
if mpi is None:
    if compiler is None:
        if (os.name != 'nt'
                and run(['which', 'mpicc'],
                        capture_output=True).returncode == 0):
            mpi = True
            compiler = 'mpicc'
        else:
            mpi = False
    elif compiler == 'mpicc':
        warn_deprecated(
            'Define in siteconfig explicitly'
            '\n\nmpi = True')
        mpi = True
    else:
        mpi = False

if mpi:
    if compiler is None:
        raise_error('Define compiler for MPI in siteconfig:'
                    "\ncompiler = ...  # MPI compiler, e.g., 'mpicc'")

# Deprecation check
if 'mpilinker' in locals():
    mpilinker = locals()['mpilinker']
    msg = ('Please remove deprecated declaration of mpilinker:'
           f'\ncompiler={repr(compiler)} will be used for linking.')
    if mpilinker == compiler:
        warn_deprecated(msg)
    else:
        msg += ('\nPlease contact GPAW developers if you need '
                'different commands for linking and compiling.')
        raise_error(msg)

# Deprecation check
for key in ['libraries', 'library_dirs', 'include_dirs',
            'runtime_library_dirs', 'define_macros']:
    mpi_key = 'mpi_' + key
    if mpi_key in locals():
        warn_deprecated(
            f'Please remove deprecated declaration of {mpi_key}'
            f' and use only {key} instead.'
            f'\nAdding {mpi_key} to {key}.')
        locals()[key] += locals()[mpi_key]


if parallel_python_interpreter:
    parallel_python_exefile = None
    if not mpi:
        raise_error('MPI is needed for parallel_python_interpreter.'
                    ' Define in siteconfig:'
                    '\n\nparallel_python_interpreter = True'
                    '\nmpi = True'
                    "\ncompiler = ...  # MPI compiler, e.g., 'mpicc'"
                    )

if mpi:
    print('Building GPAW with MPI support.')

platform_id = os.getenv('CPU_ARCH')
if platform_id:
    os.environ['_PYTHON_HOST_PLATFORM'] = get_platform() + '-' + platform_id

for flag, name in [(noblas, 'GPAW_WITHOUT_BLAS'),
                   (nolibxc, 'GPAW_WITHOUT_LIBXC'),
                   (mpi, 'PARALLEL'),
                   (fftw, 'GPAW_WITH_FFTW'),
                   (scalapack, 'GPAW_WITH_SL'),
                   (libvdwxc, 'GPAW_WITH_LIBVDWXC'),
                   (elpa, 'GPAW_WITH_ELPA')]:
    if flag and name not in [n for (n, _) in define_macros]:
        define_macros.append((name, None))

sources = [Path('c/bmgs/bmgs.c')]
sources += Path('c').glob('*.c')
sources += Path('c/xc').glob('*.c')
sources.remove(Path('c/main.c'))  # For gpaw-python executable only
if nolibxc:
    for name in ['libxc.c', 'm06l.c',
                 'tpss.c', 'revtpss.c', 'revtpss_c_pbe.c',
                 'xc_mgga.c']:
        sources.remove(Path(f'c/xc/{name}'))
    if 'xc' in libraries:
        libraries.remove('xc')

# Make build process deterministic (for "reproducible build")
sources = [str(source) for source in sources]
sources.sort()

check_dependencies(sources)

# Convert Path objects to str:
runtime_library_dirs = [str(dir) for dir in runtime_library_dirs]
library_dirs = [str(dir) for dir in library_dirs]
include_dirs = [str(dir) for dir in include_dirs]

define_macros = [(macro, value) for macro, value in define_macros
                 if macro not in undef_macros]

extensions = [Extension('_gpaw',
                        sources,
                        libraries=libraries,
                        library_dirs=library_dirs,
                        include_dirs=include_dirs,
                        define_macros=define_macros,
                        undef_macros=undef_macros,
                        extra_link_args=extra_link_args,
                        extra_compile_args=extra_compile_args,
                        runtime_library_dirs=runtime_library_dirs,
                        extra_objects=extra_objects,
                        language='c')]


if os.environ.get('GPAW_GPU'):
    # Hardcoded for LUMI right now!
    target = os.environ.get('HCC_AMDGPU_TARGET', 'gfx90a')
    # TODO: Build this also via extension
    assert os.system(
        f'HCC_AMDGPU_TARGET={target} hipcc -fPIC -fgpu-rdc '
        '-c c/gpu/hip_kernels.cpp -o c/gpu/hip_kernels.o') == 0
    assert os.system(
        f'HCC_AMDGPU_TARGET={target} hipcc -shared -fgpu-rdc --hip-link '
        '-o c/gpu/hip_kernels.so c/gpu/hip_kernels.o') == 0

    extensions.append(
        Extension('_gpaw_gpu',
                  ['c/gpu/gpaw_gpu.c'],
                  libraries=[],
                  library_dirs=['c/gpu'],
                  setup_requires=['numpy'],
                  include_dirs=include_dirs,
                  define_macros=[('NPY_NO_DEPRECATED_API', 7)],
                  undef_macros=[],
                  extra_link_args=[
                      f'-Wl,-rpath={Path("c/gpu").resolve()}'],
                  extra_compile_args=['-std=c99'],
                  # ,'-Werror=implicit-function-declaration'],
                  runtime_library_dirs=['c/gpu'],
                  extra_objects=[
                      str(Path('c/gpu/hip_kernels.so').resolve())]))


write_configuration(define_macros, include_dirs, libraries, library_dirs,
                    extra_link_args, extra_compile_args,
                    runtime_library_dirs, extra_objects, compiler)


class build_ext(_build_ext):
    def run(self):
        import numpy as np
        self.include_dirs.append(np.get_include())
        super().run()

    def build_extensions(self):
        # Override the compiler executables
        # A hack to change the used compiler and linker, inspired by
        # https://shwina.github.io/custom-compiler-linker-extensions/
        for (name, my_args) in [('compiler', compiler_args),
                                ('compiler_so', compiler_args),
                                ('linker_so', linker_so_args),
                                ('linker_exe', linker_exe_args)]:
            new_args = []
            old_args = getattr(self.compiler, name)
            # Set executable
            if compiler is not None:
                new_args += [compiler]
            else:
                new_args += [old_args[0]]
            # Set args
            if my_args is not None:
                new_args += my_args
            else:
                new_args += old_args[1:]
            self.compiler.set_executable(name, new_args)

        super().build_extensions()

        if parallel_python_interpreter:
            global parallel_python_exefile

            assert len(self.extensions) == 1, \
                'Fix gpaw-python build for multiple extensions'
            extension = self.extensions[0]

            # Path for the bin (analogous to build_lib)
            build_bin = Path(str(self.build_lib).replace('lib', 'bin'))

            # List of object files already built for the extension
            objects = []
            for src in sources:
                # Do not include _gpaw_so.o in the gpaw-python executable
                if src == 'c/_gpaw_so.c':
                    continue
                obj = Path(self.build_temp) / Path(src).with_suffix('.o')
                objects.append(str(obj))

            # Build gpaw-python
            parallel_python_exefile = build_interpreter(
                self.compiler, extension, objects,
                build_temp=self.build_temp,
                build_bin=build_bin,
                debug=self.debug)

        print("Build temp:", self.build_temp)
        print("Build lib: ", self.build_lib)
        if parallel_python_interpreter:
            print("Build bin: ", build_bin)


class install(_install):
    def run(self):
        super().run()
        if parallel_python_interpreter:
            self.copy_file(parallel_python_exefile, self.install_scripts)


class develop(_develop):
    def run(self):
        super().run()
        if parallel_python_interpreter:
            self.copy_file(parallel_python_exefile, self.script_dir)


cmdclass = {'build_ext': build_ext,
            'install': install,
            'develop': develop}

files = ['gpaw-analyse-basis', 'gpaw-basis',
         'gpaw-plot-parallel-timings', 'gpaw-runscript',
         'gpaw-setup', 'gpaw-upfplot']
scripts = [str(Path('tools') / script) for script in files]


setup(name='gpaw',
      version=version,
      description=description,
      long_description=long_description,
      maintainer='GPAW-community',
      maintainer_email='gpaw-users@listserv.fysik.dtu.dk',
      url='https://wiki.fysik.dtu.dk/gpaw',
      license='GPLv3+',
      platforms=['unix'],
      packages=find_packages(),
      package_data={'gpaw': ['py.typed']},
      entry_points={
          'console_scripts': ['gpaw = gpaw.cli.main:main']},
      setup_requires=['numpy'],
      install_requires=[f'ase>={ase_version_required}',
                        'scipy>=1.6.0',
                        'pyyaml'],
      extras_require={'docs': ['sphinx-rtd-theme',
                               'graphviz'],
                      'devel': ['flake8',
                                'mypy',
                                'pytest-xdist',
                                'interrogate']},
      ext_modules=extensions,
      scripts=scripts,
      cmdclass=cmdclass,
      classifiers=[
          'Development Status :: 6 - Mature',
          'License :: OSI Approved :: '
          'GNU General Public License v3 or later (GPLv3+)',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Topic :: Scientific/Engineering :: Physics'])



from ase.units import Bohr, Hartree

from gpaw.calculator import GPAW
from gpaw.io import Reader
from gpaw.solvation.hamiltonian import SolvationRealSpaceHamiltonian


class SolvationGPAW(GPAW):
    """Subclass of gpaw.GPAW calculator with continuum solvent model.

    See also Section III of
    A. Held and M. Walter, J. Chem. Phys. 141, 174108 (2014).
    """

    def __init__(self, restart=None, cavity=None, dielectric=None,
                 interactions=None, **gpaw_kwargs):
        """Constructor for SolvationGPAW class.

        Additional arguments not present in GPAW class:
        cavity       -- A Cavity instance.
        dielectric   -- A Dielectric instance.
        interactions -- A list of Interaction instances.
        """
        if interactions is None:
            interactions = []

        # if not all([cavity, dielectric]):
        #    raise IOError('Cavity and dielectric modules need to be '
        #                  'defined in the calculator')

        self.stuff_for_hamiltonian = (cavity, dielectric, interactions)

        GPAW.__init__(self, restart, **gpaw_kwargs)

    def read(self, filename):
        """Read yourself from a file"""
        self.reader = reader = Reader(filename)
        if 'implicit_solvent' in reader:
            impl_in = reader.implicit_solvent
            if 'name' in impl_in.cavity.effective_potential:
                efpot = impl_in.cavity.effective_potential

                def atomic_radii(atoms):
                    return efpot.atomic_radii

                if efpot.name == 'SJMPower12Potential':
                    from gpaw.solvation.sjm import SJMPower12Potential

                    effective_potential = SJMPower12Potential(
                        atomic_radii=atomic_radii,
                        u0=efpot.u0,
                        H2O_layer=efpot.H2O_layer,
                        unsolv_backside=efpot.unsolv_backside)
                elif efpot.name == 'Power12Potential':
                    from gpaw.solvation.sjm import Power12Potential
                    effective_potential = Power12Potential(
                        atomic_radii=atomic_radii,
                        u0=efpot.u0)
                else:
                    raise IOError('Reading the given effective potential '
                                  'is not implemented yet')

            if 'name' in impl_in.cavity.surface_calculator:
                suca = impl_in.cavity.surface_calculator
                if suca.name == 'GradientSurface':
                    from gpaw.solvation.cavity import GradientSurface
                    surface_calculator = GradientSurface(suca.nn)
                else:
                    raise IOError('Reading in the given used surface '
                                  'calculator is not implemented')

            T = impl_in.cavity.temperature

            from gpaw.solvation.cavity import EffectivePotentialCavity
            cavity = EffectivePotentialCavity(
                effective_potential=effective_potential,
                temperature=T,
                surface_calculator=surface_calculator)

            if impl_in.dielectric.name == 'LinearDielectric':
                from gpaw.solvation.dielectric import LinearDielectric
                dielectric = LinearDielectric(epsinf=impl_in.dielectric.epsinf)

            if impl_in.interactions.name == 'SurfaceInteraction':
                suin = impl_in.interactions
                from gpaw.solvation.interactions import SurfaceInteraction
                interactions = [SurfaceInteraction(suin.surface_tension)]

            self.stuff_for_hamiltonian = (cavity, dielectric, interactions)

        reader = GPAW.read(self, filename)
        return reader

    def _write(self, writer, mode):
        GPAW._write(self, writer, mode)
        stuff = self.stuff_for_hamiltonian
        writer.child('implicit_solvent').write(cavity=stuff[0],
                                               dielectric=stuff[1],
                                               interactions=stuff[2][0])

    def create_hamiltonian(self, realspace, mode, xc):
        if not realspace:
            raise NotImplementedError(
                'SolvationGPAW does not support '
                'calculations in reciprocal space yet.')

        dens = self.density
        self.hamiltonian = SolvationRealSpaceHamiltonian(
            *self.stuff_for_hamiltonian,
            gd=dens.gd, finegd=dens.finegd,
            nspins=dens.nspins,
            collinear=dens.collinear,
            setups=dens.setups,
            timer=self.timer,
            xc=xc,
            world=self.world,
            redistributor=dens.redistributor,
            vext=self.parameters.external,
            psolver=self.parameters.poissonsolver,
            stencil=mode.interpolation)

        self.log(self.hamiltonian)

        xc.set_grid_descriptor(self.hamiltonian.finegd)

    def initialize_positions(self, atoms=None):
        spos_ac = GPAW.initialize_positions(self, atoms)
        self.hamiltonian.update_atoms(self.atoms, self.log)
        return spos_ac

    def get_electrostatic_energy(self):
        """Return electrostatic part of the total energy.

        The electrostatic part consists of everything except
        the short-range interactions defined in the interactions list.
        """
        # Energy extrapolated to zero width:
        return Hartree * self.hamiltonian.e_el_extrapolated

    def get_solvation_interaction_energy(self, subscript, atoms=None):
        """Return a specific part of the solvation interaction energy.

        The subscript parameter defines which part is to be returned.
        It has to match the value of a subscript attribute of one of
        the interactions in the interactions list.
        """
        return Hartree * getattr(self.hamiltonian, 'e_' + subscript)

    def get_cavity_volume(self, atoms=None):
        """Return the cavity volume in Angstrom ** 3.

        In case no volume calculator has been set for the cavity, None
        is returned.
        """
        V = self.hamiltonian.cavity.V
        return V and V * Bohr ** 3

    def get_cavity_surface(self, atoms=None):
        """Return the cavity surface area in Angstrom ** 2.

        In case no surface calculator has been set for the cavity,
        None is returned.
        """
        A = self.hamiltonian.cavity.A
        return A and A * Bohr ** 2



import pytest
from ase.build import molecule

from gpaw import GPAW, FermiDirac
from gpaw.cluster import Cluster
from gpaw.lrtddft import LrTDDFT


@pytest.mark.lrtddft
def test_lrtddft(in_tmp_dir):
    from ase.vibrations.resonant_raman import ResonantRamanCalculator
    h = 0.25
    H2 = Cluster(molecule('H2'))
    H2.minimal_box(3., h=h)
    H2.calc = GPAW(h=h, occupations=FermiDirac(width=0.2),
                   symmetry='off')

    rr = ResonantRamanCalculator(
        H2, LrTDDFT, exkwargs={'restrict': {'energy_range': 15}})
    rr.run()



# fails with On entry to ZGEMV parameter number 8 had an illegal value

import pytest
from gpaw.mpi import world
from ase.build import molecule
from gpaw import GPAW
from gpaw import PW

pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


def test_pw_moleculecg():
    m = molecule('H')
    m.center(vacuum=2.0)
    m.calc = GPAW(mode=PW(), eigensolver='cg')
    m.get_potential_energy()



"""
Python wrapper for FFTW3 library
================================

.. autoclass:: FFTPlans

"""
from __future__ import annotations

import weakref
from types import ModuleType

import numpy as np
from scipy.fft import fftn, ifftn, irfftn, rfftn

import _gpaw
from gpaw import SCIPY_VERSION
from gpaw.typing import Array1D, Array3D, DTypeLike, IntVector

ESTIMATE = 64
MEASURE = 0
PATIENT = 32
EXHAUSTIVE = 8

_plan_cache: dict[tuple, weakref.ReferenceType] = {}


def have_fftw() -> bool:
    """Did we compile with FFTW?"""
    return hasattr(_gpaw, 'FFTWPlan')


def check_fft_size(n: int, factors=[2, 3, 5, 7]) -> bool:
    """Check if n is an efficient fft size.

    Efficient means that n can be factored into small primes (2, 3, 5, 7).

    >>> check_fft_size(17)
    False
    >>> check_fft_size(18)
    True
    """

    if n == 1:
        return True
    for x in factors:
        if n % x == 0:
            return check_fft_size(n // x, factors)
    return False


def get_efficient_fft_size(N: int, n=1, factors=[2, 3, 5, 7]) -> int:
    """Return smallest efficient fft size.

    Must be greater than or equal to N and divisible by n.

    >>> get_efficient_fft_size(17)
    18
    """
    N = -(-N // n) * n
    while not check_fft_size(N, factors):
        N += n
    return N


def empty(shape, dtype=float):
    """numpy.empty() equivalent with 16 byte alignment."""
    assert dtype == complex
    N = np.prod(shape)
    a = np.empty(2 * N + 1)
    offset = (a.ctypes.data % 16) // 8
    a = a[offset:2 * N + offset].view(complex)
    a.shape = shape
    return a


def create_plans(size_c: IntVector,
                 dtype: DTypeLike,
                 flags: int = MEASURE,
                 xp: ModuleType = np) -> FFTPlans:
    """Create plan-objects for FFT and inverse FFT."""
    key = (tuple(size_c), dtype, flags, xp)
    # Look up weakref to plan:
    if key in _plan_cache:
        plan = _plan_cache[key]()
        # Check if plan is still "alive":
        if plan is not None:
            return plan
    # Create new plan:
    if xp is not np:
        plan = CuPyFFTPlans(size_c, dtype)
    elif have_fftw():
        plan = FFTWPlans(size_c, dtype, flags)
    else:
        plan = NumpyFFTPlans(size_c, dtype)
    _plan_cache[key] = weakref.ref(plan)
    return plan


class FFTPlans:
    def __init__(self,
                 size_c: IntVector,
                 dtype: DTypeLike,
                 empty=empty):
        self.shape: tuple[int, ...]
        if dtype == float:
            self.shape = (size_c[0], size_c[1], size_c[2] // 2 + 1)
            self.tmp_Q = empty(self.shape, complex)
            self.tmp_R = self.tmp_Q.view(float)[:, :, :size_c[2]]
        else:
            self.shape = tuple(size_c)
            self.tmp_Q = empty(size_c, complex)
            self.tmp_R = self.tmp_Q

    def fft(self) -> None:
        """Do FFT from ``tmp_R`` to ``tmp_Q``.

        >>> plans = create_plans([4, 1, 1], float)
        >>> plans.tmp_R[:, 0, 0] = [1, 0, 1, 0]
        >>> plans.fft()
        >>> plans.tmp_Q[:, 0, 0]
        array([2.+0.j, 0.+0.j, 2.+0.j, 0.+0.j])
        """
        raise NotImplementedError

    def ifft(self) -> None:
        """Do inverse FFT from ``tmp_Q`` to ``tmp_R``.

        >>> plans = create_plans([4, 1, 1], complex)
        >>> plans.tmp_Q[:, 0, 0] = [0, 1j, 0, 0]
        >>> plans.ifft()
        >>> plans.tmp_R[:, 0, 0]
        array([ 0.+1.j, -1.+0.j,  0.-1.j,  1.+0.j])
        """
        raise NotImplementedError

    def ifft_sphere(self, coef_G, pw, out_R):
        if coef_G is None:
            out_R.scatter_from(None)
            return
        pw.paste(coef_G, self.tmp_Q)
        if pw.dtype == float:
            t = self.tmp_Q[:, :, 0]
            n, m = (s // 2 - 1 for s in out_R.desc.size_c[:2])
            t[0, -m:] = t[0, m:0:-1].conj()
            t[n:0:-1, -m:] = t[-n:, m:0:-1].conj()
            t[-n:, -m:] = t[n:0:-1, m:0:-1].conj()
            t[-n:, 0] = t[n:0:-1, 0].conj()
        self.ifft()
        out_R.scatter_from(self.tmp_R)

    def fft_sphere(self, in_R, pw):
        self.tmp_R[:] = in_R.data
        self.fft()
        coefs = pw.cut(self.tmp_Q) * (1 / self.tmp_R.size)
        return coefs


class FFTWPlans(FFTPlans):
    """FFTW3 3d transforms."""
    def __init__(self, size_c, dtype, flags=MEASURE):
        if not have_fftw():
            raise ImportError('Not compiled with FFTW.')
        super().__init__(size_c, dtype)
        self._fftplan = _gpaw.FFTWPlan(self.tmp_R, self.tmp_Q, -1, flags)
        self._ifftplan = _gpaw.FFTWPlan(self.tmp_Q, self.tmp_R, 1, flags)

    def fft(self):
        _gpaw.FFTWExecute(self._fftplan)

    def ifft(self):
        _gpaw.FFTWExecute(self._ifftplan)

    def __del__(self):
        _gpaw.FFTWDestroy(self._fftplan)
        _gpaw.FFTWDestroy(self._ifftplan)


class NumpyFFTPlans(FFTPlans):
    """Numpy fallback."""
    def fft(self):
        if self.tmp_R.dtype == float:
            self.tmp_Q[:] = rfftn(self.tmp_R, overwrite_x=True)
        else:
            self.tmp_Q[:] = fftn(self.tmp_R, overwrite_x=True)

    def ifft(self):
        if self.tmp_R.dtype == float:
            self.tmp_R[:] = irfftn(self.tmp_Q, self.tmp_R.shape,
                                   norm='forward', overwrite_x=True)
        else:
            self.tmp_R[:] = ifftn(self.tmp_Q, self.tmp_R.shape,
                                  norm='forward', overwrite_x=True)


class CuPyFFTPlans(FFTPlans):
    def __init__(self,
                 size_c: IntVector,
                 dtype: DTypeLike):
        from gpaw.core.plane_waves import PlaneWaves
        from gpaw.gpu import cupy as cp
        self.dtype = dtype
        super().__init__(size_c, dtype, empty=cp.empty)
        self.Q_G_cache: dict[PlaneWaves, Array1D] = {}

    def fft(self):
        from gpaw.gpu import cupyx
        if self.tmp_R.dtype == float:
            self.tmp_Q[:] = cupyx.scipy.fft.rfftn(self.tmp_R)
        else:
            self.tmp_Q[:] = cupyx.scipy.fft.fftn(self.tmp_R)

    def ifft(self):
        from gpaw.gpu import cupyx
        if self.tmp_R.dtype == float:
            self.tmp_R[:] = cupyx.scipy.fft.irfftn(
                self.tmp_Q, self.tmp_R.shape,
                norm='forward',
                overwrite_x=True)
        else:
            self.tmp_R[:] = cupyx.scipy.fft.ifftn(
                self.tmp_Q, self.tmp_R.shape,
                norm='forward',
                overwrite_x=True)

    def indices(self, pw):
        from gpaw.gpu import cupy as cp
        Q_G = self.Q_G_cache.get(pw)
        if Q_G is None:
            Q_G = cp.asarray(pw.indices(self.shape))
            self.Q_G_cache[pw] = Q_G
        return Q_G

    def ifft_sphere(self, coef_G, pw, out_R):
        from gpaw.gpu import cupyx

        if coef_G is None:
            out_R.scatter_from(None)
            return

        if out_R.desc.comm.size == 1:
            array_R = out_R.data
        else:
            array_R = self.tmp_R
        array_Q = self.tmp_Q

        array_Q[:] = 0.0
        Q_G = self.indices(pw)
        array_Q.ravel()[Q_G] = coef_G

        assert SCIPY_VERSION >= [1, 6]
        if self.dtype == complex:
            array_R[:] = cupyx.scipy.fft.ifftn(
                array_Q, array_Q.shape,
                norm='forward', overwrite_x=True)
        else:
            # We need a GPU kernel for this stuff:
            t = array_Q[:, :, 0]
            n, m = (s // 2 - 1 for s in out_R.desc.size_c[:2])
            t[0, -m:] = t[0, m:0:-1].conj()
            t[n:0:-1, -m:] = t[-n:, m:0:-1].conj()
            t[-n:, -m:] = t[n:0:-1, m:0:-1].conj()
            t[-n:, 0] = t[n:0:-1, 0].conj()
            array_R[:] = cupyx.scipy.fft.irfftn(
                array_Q, out_R.desc.global_shape(),
                norm='forward', overwrite_x=True)

        if out_R.desc.comm.size > 1:
            out_R.scatter_from(array_R)

    def fft_sphere(self, in_R, pw):
        from gpaw.gpu import cupyx
        if self.dtype == complex:
            out_Q = cupyx.scipy.fft.fftn(in_R)
        else:
            out_Q = cupyx.scipy.fft.rfftn(in_R)
        Q_G = self.indices(pw)
        coef_G = out_Q.ravel()[Q_G] * (1 / in_R.size)
        return coef_G


# The rest of this file will be removed in the future ...

def check_fftw_inputs(in_R, out_R):
    for arr in in_R, out_R:
        # Note: Arrays not necessarily contiguous due to 16-byte alignment
        assert arr.ndim == 3  # We can perhaps relax this requirement
        assert arr.dtype == float or arr.dtype == complex

    if in_R.dtype == out_R.dtype == complex:
        assert in_R.shape == out_R.shape
    else:
        # One real and one complex:
        R, C = (in_R, out_R) if in_R.dtype == float else (out_R, in_R)
        assert C.dtype == complex
        assert R.shape[:2] == C.shape[:2]
        assert C.shape[2] == 1 + R.shape[2] // 2


class FFTPlan:
    """FFT 3d transform."""
    def __init__(self,
                 in_R: Array3D,
                 out_R: Array3D,
                 sign: int,
                 flags: int = MEASURE):
        check_fftw_inputs(in_R, out_R)
        self.in_R = in_R
        self.out_R = out_R
        self.sign = sign
        self.flags = flags

    def execute(self) -> None:
        raise NotImplementedError


class FFTWPlan(FFTPlan):
    """FFTW3 3d transform."""
    def __init__(self, in_R, out_R, sign, flags=MEASURE):
        if not have_fftw():
            raise ImportError('Not compiled with FFTW.')
        self._ptr = _gpaw.FFTWPlan(in_R, out_R, sign, flags)
        FFTPlan.__init__(self, in_R, out_R, sign, flags)

    def execute(self):
        _gpaw.FFTWExecute(self._ptr)

    def __del__(self):
        if getattr(self, '_ptr', None) and _gpaw is not None:
            _gpaw.FFTWDestroy(self._ptr)
        self._ptr = None


class NumpyFFTPlan(FFTPlan):
    """Numpy fallback."""
    def execute(self):
        if self.in_R.dtype == float:
            self.out_R[:] = np.fft.rfftn(self.in_R)
        elif self.out_R.dtype == float:
            self.out_R[:] = np.fft.irfftn(self.in_R, self.out_R.shape)
            self.out_R *= self.out_R.size
        elif self.sign == 1:
            self.out_R[:] = np.fft.ifftn(self.in_R, self.out_R.shape)
            self.out_R *= self.out_R.size
        else:
            self.out_R[:] = np.fft.fftn(self.in_R)


def create_plan(in_R: Array3D,
                out_R: Array3D,
                sign: int,
                flags: int = MEASURE) -> FFTPlan:
    if have_fftw():
        return FFTWPlan(in_R, out_R, sign, flags)
    return NumpyFFTPlan(in_R, out_R, sign, flags)



from math import pi, sqrt
from warnings import warn
from typing import Optional, List, Dict

import numpy as np
from ase.units import Bohr, Ha

from gpaw.calculator import GPAW
from gpaw.atom.shapefunc import shape_functions
from gpaw.fftw import get_efficient_fft_size
from gpaw.grid_descriptor import GridDescriptor
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.utilities import h2gpts
from gpaw.pw.descriptor import PWDescriptor
from gpaw.mpi import serial_comm
from gpaw.setup import Setup
from gpaw.spline import Spline
from gpaw.typing import Array3D


class Interpolator:
    def __init__(self, gd1, gd2, dtype=float):
        self.pd1 = PWDescriptor(0.0, gd1, dtype)
        self.pd2 = PWDescriptor(0.0, gd2, dtype)

    def interpolate(self, a_r):
        return self.pd1.interpolate(a_r, self.pd2)[0]


POINTS = 200


class PS2AE:
    """Transform PS to AE wave functions.

    Interpolates PS wave functions to a fine grid and adds PAW
    corrections in order to obtain true AE wave functions.
    """
    def __init__(self,
                 calc: GPAW,
                 grid_spacing: float = 0.05,
                 n: int = 2,
                 h=None  # deprecated
                 ):
        """Create transformation object.

        calc: GPAW calculator object
            The calcalator that has the wave functions.
        grid_spacing: float
            Desired grid-spacing in Angstrom.
        n: int
            Force number of points to be a mulitiple of n.
        """
        if h is not None:
            warn('Please use grid_spacing=... instead of h=...')
            grid_spacing = h

        self.calc = calc
        gd = calc.wfs.gd

        gd1 = GridDescriptor(gd.N_c, gd.cell_cv, comm=serial_comm)

        # Descriptor for the final grid:
        N_c = h2gpts(grid_spacing / Bohr, gd.cell_cv)
        N_c = np.array([get_efficient_fft_size(N, n) for N in N_c])
        gd2 = self.gd = GridDescriptor(N_c, gd.cell_cv, comm=serial_comm)
        self.interpolator = Interpolator(gd1, gd2, self.calc.wfs.dtype)

        self._dphi: Optional[LFC] = None  # PAW correction

        self.dv = self.gd.dv * Bohr**3

    @property
    def dphi(self) -> LFC:
        if self._dphi is not None:
            return self._dphi

        splines: Dict[Setup, List[Spline]] = {}
        dphi_aj = []
        for setup in self.calc.wfs.setups:
            dphi_j = splines.get(setup)
            if dphi_j is None:
                rcut = max(setup.rcut_j) * 1.1
                gcut = setup.rgd.ceil(rcut)
                dphi_j = []
                for l, phi_g, phit_g in zip(setup.l_j,
                                            setup.data.phi_jg,
                                            setup.data.phit_jg):
                    dphi_g = (phi_g - phit_g)[:gcut]
                    dphi_j.append(setup.rgd.spline(dphi_g, rcut, l,
                                                   points=200))
                splines[setup] = dphi_j
            dphi_aj.append(dphi_j)

        self._dphi = LFC(self.gd, dphi_aj, kd=self.calc.wfs.kd.copy(),
                         dtype=self.calc.wfs.dtype)
        self._dphi.set_positions(self.calc.spos_ac)

        return self._dphi

    def get_wave_function(self,
                          n: int,
                          k: int = 0,
                          s: int = 0,
                          ae: bool = True,
                          periodic: bool = False) -> Array3D:
        """Interpolate wave function.

        Returns 3-d array in units of Ang**-1.5.

        n: int
            Band index.
        k: int
            K-point index.
        s: int
            Spin index.
        ae: bool
            Add PAW correction to get an all-electron wave function.
        periodic:
            Return periodic part of wave-function, u(r), instead of
            psi(r)=exp(ikr)u(r).
        """
        u_r = self.calc.get_pseudo_wave_function(n, k, s,
                                                 periodic=True)
        u_R = self.interpolator.interpolate(u_r * Bohr**1.5)

        k_c = self.calc.wfs.kd.ibzk_kc[k]
        gamma = np.isclose(k_c, 0.0).all()

        if gamma:
            eikr_R = 1.0
        else:
            eikr_R = self.gd.plane_wave(k_c)

        if ae:
            dphi = self.dphi
            wfs = self.calc.wfs
            P_nI = wfs.collect_projections(k, s)

            if wfs.world.rank == 0:
                psi_R = u_R * eikr_R
                P_ai = {}
                I1 = 0
                for a, setup in enumerate(wfs.setups):
                    I2 = I1 + setup.ni
                    P_ai[a] = P_nI[n, I1:I2]
                    I1 = I2
                dphi.add(psi_R, P_ai, k)
                u_R = psi_R / eikr_R

            wfs.world.broadcast(u_R, 0)

        if periodic:
            return u_R * Bohr**-1.5
        else:
            return u_R * eikr_R * Bohr**-1.5

    def get_pseudo_density(self,
                           add_compensation_charges: bool = True) -> Array3D:
        """Interpolate pseudo density."""
        dens = self.calc.density
        gd1 = dens.gd
        assert gd1.comm.size == 1
        interpolator = Interpolator(gd1, self.gd)
        dens_r = dens.nt_sG[:dens.nspins].sum(axis=0)
        dens_R = interpolator.interpolate(dens_r)

        if add_compensation_charges:
            dens.calculate_multipole_moments()
            ghat = LFC(self.gd, [setup.ghat_l for setup in dens.setups],
                       integral=sqrt(4 * pi))
            ghat.set_positions(self.calc.spos_ac)
            Q_aL = {}
            for a, Q_L in dens.Q_aL.items():
                Q_aL[a] = Q_L.copy()
                Q_aL[a][0] += dens.setups[a].Nv / (4 * pi)**0.5
            ghat.add(dens_R, Q_aL)

        return dens_R / Bohr**3

    def get_electrostatic_potential(self,
                                    ae: bool = True,
                                    rcgauss: float = 0.02) -> Array3D:
        """Interpolate electrostatic potential.

        Return value in eV.

        ae: bool
            Add PAW correction to get the all-electron potential.
        rcgauss: float
            Width of gaussian (in Angstrom) used to represent the nuclear
            charge.
        """
        gd = self.calc.hamiltonian.finegd
        v_r = self.calc.get_electrostatic_potential() / Ha
        gd1 = GridDescriptor(gd.N_c, gd.cell_cv, comm=serial_comm)
        interpolator = Interpolator(gd1, self.gd)
        v_R = interpolator.interpolate(v_r)

        if ae:
            self.add_potential_correction(v_R, rcgauss / Bohr)

        return v_R * Ha

    def add_potential_correction(self,
                                 v_R: Array3D,
                                 rcgauss: float) -> None:
        dens = self.calc.density
        dens.D_asp.redistribute(dens.atom_partition.as_serial())
        dens.Q_aL.redistribute(dens.atom_partition.as_serial())

        dv_a1 = []
        for a, D_sp in dens.D_asp.items():
            setup = dens.setups[a]
            c = setup.xc_correction
            rgd = c.rgd
            params = setup.data.shape_function.copy()
            params['lmax'] = 0
            ghat_g = shape_functions(rgd, **params)[0]
            Z_g = shape_functions(rgd, 'gauss', rcgauss, lmax=0)[0] * setup.Z
            D_q = np.dot(D_sp.sum(0), c.B_pqL[:, :, 0])
            dn_g = np.dot(D_q, (c.n_qg - c.nt_qg)) * sqrt(4 * pi)
            dn_g += 4 * pi * (c.nc_g - c.nct_g)
            dn_g -= Z_g
            dn_g -= dens.Q_aL[a][0] * ghat_g * sqrt(4 * pi)
            dv_g = rgd.poisson(dn_g) / sqrt(4 * pi)
            dv_g[1:] /= rgd.r_g[1:]
            dv_g[0] = dv_g[1]
            dv_g[-1] = 0.0
            dv_a1.append([rgd.spline(dv_g, points=POINTS)])

        dens.D_asp.redistribute(dens.atom_partition)
        dens.Q_aL.redistribute(dens.atom_partition)

        if dv_a1:
            dv = LFC(self.gd, dv_a1)
            dv.set_positions(self.calc.spos_ac)
            dv.add(v_R)
        dens.gd.comm.broadcast(v_R, 0)


def interpolate_weight(calc, weight, h=0.05, n=2):
    """interpolates cdft weight function, gd is the fine grid."""
    gd = calc.density.finegd

    weight = gd.collect(weight, broadcast=True)
    weight = gd.zero_pad(weight)

    w = np.zeros_like(weight)
    gd1 = GridDescriptor(gd.N_c, gd.cell_cv, comm=serial_comm)
    gd1.distribute(weight, w)

    N_c = h2gpts(h / Bohr, gd.cell_cv)
    N_c = np.array([get_efficient_fft_size(N, n) for N in N_c])
    gd2 = GridDescriptor(N_c, gd.cell_cv, comm=serial_comm)

    interpolator = Interpolator(gd1, gd2)
    W = interpolator.interpolate(w)

    return W



from gpaw import GPAW


def test_electrostatic_potential(gpw_files):
    """Make sure the whole array is returned also when parallelizing
    over plane-wave coefficients."""
    calc = GPAW(gpw_files['h2_pw'])
    v = calc.get_electrostatic_potential()
    assert v.shape == tuple(calc.hamiltonian.finegd.N_c)



import pytest
import numpy as np
import ase.io.ulm as ulm
from ase.build import molecule

from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lcaotddft.densitymatrix import DensityMatrix
from gpaw.lcaotddft.frequencydensitymatrix import FrequencyDensityMatrix
from gpaw.mpi import world
from gpaw.tddft.folding import frequencies
from gpaw.test import equal


@pytest.mark.gllb
@pytest.mark.libxc
@pytest.mark.rttddft
def test_lcaotddft_restart(in_tmp_dir):
    atoms = molecule('SiH4')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=7, h=0.4,
                basis='dzp', mode='lcao',
                convergence={'density': 1e-8},
                xc='GLLBSC',
                txt='gs.out')
    atoms.calc = calc
    _ = atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation
    td_calc = LCAOTDDFT('gs.gpw', txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    dmat = DensityMatrix(td_calc)
    freqs = frequencies(np.arange(0.05, 6.01, 1.0), None, None)
    fdm = FrequencyDensityMatrix(td_calc, dmat, frequencies=freqs)
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 3)

    # Write a restart point
    td_calc.write('td.gpw', mode='all')
    fdm.write('fdm_restart_point.ulm')

    # Keep propagating
    td_calc.propagate(20, 3)
    fdm.write('fdm_final_ref.ulm')

    # Restart from the restart point
    td_calc = LCAOTDDFT('td.gpw', txt='td2.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    dmat = DensityMatrix(td_calc)
    fdm = FrequencyDensityMatrix(td_calc, dmat,
                                 filename='fdm_restart_point.ulm')
    td_calc.propagate(20, 3)
    fdm.write('fdm_final_check.ulm')
    world.barrier()

    # Check dipole moment file
    data_tj = np.loadtxt('dm.dat')
    # Original run
    ref_i = data_tj[4:8].ravel()
    # Restarted steps
    data_i = data_tj[8:].ravel()

    tol = 1e-10
    equal(data_i, ref_i, tol)

    tol = 1e-8
    # Check frequency density matrix file
    with ulm.open('fdm_final_ref.ulm') as fdm_ref:
        with ulm.open('fdm_final_check.ulm') as fdm_check:
            for key in ['FReDrho_wuMM', 'FImDrho_wuMM', 'rho0_uMM', 'time']:
                equal(fdm_ref.get(key), fdm_check.get(key), tol)

    # Test the absolute values
    data = np.loadtxt('dm.dat')[:8].ravel()
    if 0:
        from gpaw.test import print_reference
        print_reference(data, 'ref', '%.12le')

    ref = [0.000000000000e+00,
           1.440746980000e-15,
           -5.150207058975e-14,
           -2.111502433286e-14,
           -7.898943127163e-15,
           0.000000000000e+00,
           2.611197480000e-15,
           -8.396549188150e-14,
           -2.905622138206e-14,
           -3.511635310469e-14,
           8.268274700000e-01,
           7.301260440000e-17,
           6.205222369795e-05,
           6.205222375484e-05,
           6.205222374486e-05,
           1.653654930000e+00,
           -2.352185940000e-15,
           1.001902218476e-04,
           1.001902218874e-04,
           1.001902218743e-04,
           2.480482400000e+00,
           1.199595930000e-15,
           1.069904191357e-04,
           1.069904191591e-04,
           1.069904191494e-04,
           3.307309870000e+00,
           1.993145110000e-16,
           9.190706194380e-05,
           9.190706194210e-05,
           9.190706193849e-05,
           4.134137330000e+00,
           1.527615730000e-15,
           6.808273775138e-05,
           6.808273772079e-05,
           6.808273772212e-05,
           4.960964800000e+00,
           1.490415960000e-15,
           4.135613094036e-05,
           4.135613089495e-05,
           4.135613089985e-05]

    print('result')
    print(data.tolist())

    tol = 1e-9
    equal(data, ref, tol)



from pathlib import Path
from types import SimpleNamespace

import numpy as np

from ase.units import Ha, Bohr
from ase.utils import lazyproperty

import gpaw.mpi as mpi
from gpaw.response.ibz2bz import IBZ2BZMaps


class ResponseGroundStateAdapter:
    def __init__(self, calc):
        wfs = calc.wfs

        self.kd = wfs.kd
        self.world = calc.world
        self.gd = wfs.gd
        self.finegd = calc.density.finegd
        self.bd = wfs.bd
        self.nspins = wfs.nspins
        self.dtype = wfs.dtype

        self.spos_ac = calc.spos_ac

        self.kpt_u = wfs.kpt_u
        self.kpt_qs = wfs.kpt_qs

        self.fermi_level = wfs.fermi_level
        self.atoms = calc.atoms
        self.pawdatasets = [ResponsePAWDataset(setup) for setup in calc.setups]

        self.pbc = self.atoms.pbc
        self.volume = self.gd.volume

        self.nvalence = wfs.nvalence

        self.ibz2bz = IBZ2BZMaps.from_calculator(calc)

        self._wfs = wfs
        self._density = calc.density
        self._hamiltonian = calc.hamiltonian
        self._calc = calc

    @classmethod
    def from_gpw_file(cls, gpw, context):
        """Initiate the ground state adapter directly from a .gpw file."""
        from gpaw import GPAW, disable_dry_run
        assert Path(gpw).is_file()

        context.print('Reading ground state calculation:\n  %s' % gpw)

        with context.timer('Read ground state'):
            with disable_dry_run():
                calc = GPAW(gpw, txt=None, communicator=mpi.serial_comm)

        return cls(calc)

    @property
    def pd(self):
        # This is an attribute error in FD/LCAO mode.
        # We need to abstract away "calc" in all places used by response
        # code, and that includes places that are also compatible with FD.
        return self._wfs.pd

    @lazyproperty
    def global_pd(self):
        """Get a PWDescriptor that includes all k-points.

        In particular, this is necessary to allow all cores to be able to work
        on all k-points in the case where calc is parallelized over k-points,
        see gpaw.response.kspair
        """
        from gpaw.pw.descriptor import PWDescriptor

        assert self.gd.comm.size == 1
        kd = self.kd.copy()  # global KPointDescriptor without a comm
        return PWDescriptor(self.pd.ecut, self.gd,
                            dtype=self.pd.dtype,
                            kd=kd, fftwflags=self.pd.fftwflags,
                            gammacentered=self.pd.gammacentered)

    def get_occupations_width(self):
        # Ugly hack only used by pair.intraband_pair_density I think.
        # Actually: was copy-pasted in chi0 also.
        # More duplication can probably be eliminated around those.

        # Only works with Fermi-Dirac distribution
        occs = self._wfs.occupations
        assert occs.name in {'fermi-dirac', 'zero-width'}

        # No carriers when T=0
        width = getattr(occs, '_width', 0.0) / Ha
        return width

    def nonpbc_cell_product(self):
        """Volume, area, or length, taken in all non-periodic directions."""
        nonpbc = ~self.pbc
        cell_cv = self.gd.cell_cv
        return abs(np.linalg.det(cell_cv[nonpbc][:, nonpbc]))

    @property
    def nt_sR(self):
        # Used by localft and fxc_kernels
        return self._density.nt_sG

    @property
    def nt_sr(self):
        # Used by localft
        if self._density.nt_sg is None:
            self._density.interpolate_pseudo_density()
        return self._density.nt_sg

    @property
    def D_asp(self):
        # Used by fxc_kernels
        return self._density.D_asp

    def get_pseudo_density(self, gridrefinement=2):
        # Used by localft
        if gridrefinement == 1:
            return self.nt_sR, self.gd
        elif gridrefinement == 2:
            return self.nt_sr, self.finegd
        else:
            raise ValueError(f'Invalid gridrefinement {gridrefinement}')

    def get_all_electron_density(self, gridrefinement=2):
        # Used by fxc, fxc_kernels and localft
        return self._density.get_all_electron_density(
            atoms=self.atoms, gridrefinement=gridrefinement)

    # Things used by EXX.  This is getting pretty involved.
    #
    # EXX naughtily accesses the density object in order to
    # interpolate_pseudo_density() which is in principle mutable.

    def hacky_all_electron_density(self, **kwargs):
        # fxc likes to get all electron densities.  It calls
        # calc.get_all_electron_density() and so we wrap that here.
        # But it also collects to serial (bad), and it also zeropads
        # nonperiodic directions (probably WRONG!).
        #
        # Also this one returns in user units, whereas the calling
        # code actually wants internal units.  Very silly then.
        #
        # ALso, the calling code often wants the gd, which is not
        # returned, so it is redundantly reconstructed in multiple
        # places by refining the "right" number of times.
        n_g = self._calc.get_all_electron_density(**kwargs)
        n_g *= Bohr**3
        return n_g

    # Used by EXX.
    @property
    def hamiltonian(self):
        return self._hamiltonian

    # Used by EXX.
    @property
    def density(self):
        return self._density

    # Ugh SOC
    def soc_eigenstates(self, **kwargs):
        from gpaw.spinorbit import soc_eigenstates
        return soc_eigenstates(self._calc, **kwargs)

    @property
    def xcname(self):
        return self.hamiltonian.xc.name

    def get_xc_difference(self, xc):
        # XXX used by gpaw/xc/tools.py
        return self._calc.get_xc_difference(xc)

    def get_wave_function_array(self, u, n):
        # XXX used by gpaw/xc/tools.py in a hacky way
        return self._wfs._get_wave_function_array(
            u, n, realspace=True)

    def pair_density_paw_corrections(self, qpd):
        from gpaw.response.paw import get_pair_density_paw_corrections
        return get_pair_density_paw_corrections(
            pawdatasets=self.pawdatasets, qpd=qpd, spos_ac=self.spos_ac)

    def get_pos_av(self):
        # gd.cell_cv must always be the same as pd.gd.cell_cv, right??
        return np.dot(self.spos_ac, self.gd.cell_cv)

    def count_occupied_bands(self, ftol=1e-6):
        """Count the number of filled and non-empty bands.

        ftol : float
            Threshold determining whether a band is completely filled
            (f > 1 - ftol) or completely empty (f < ftol).
        """
        nocc1 = 9999999
        nocc2 = 0
        for kpt in self.kpt_u:
            f_n = kpt.f_n / kpt.weight
            nocc1 = min((f_n > 1 - ftol).sum(), nocc1)
            nocc2 = max((f_n > ftol).sum(), nocc2)
        return nocc1, nocc2

    @property
    def ibzq_qc(self):
        # For G0W0Kernel
        kd = self.kd
        bzq_qc = kd.get_bz_q_points(first=True)
        U_scc = kd.symmetry.op_scc
        ibzq_qc = kd.get_ibz_q_points(bzq_qc, U_scc)[0]

        return ibzq_qc

    def get_ibz_vertices(self):
        # For the tetrahedron method in Chi0
        from gpaw.bztools import get_bz
        # NB: We are ignoring the pbc_c keyword to get_bz() in order to mimic
        # find_high_symmetry_monkhorst_pack() in gpaw.bztools. XXX
        _, ibz_vertices_kc = get_bz(self._calc)
        return ibz_vertices_kc


# Contains all the relevant information
# from Setups class for response calculators
class ResponsePAWDataset:
    def __init__(self, setup):
        self.ni = setup.ni
        self.rgd = setup.rgd
        self.rcut_j = setup.rcut_j
        self.l_j = setup.l_j
        self.lq = setup.lq
        self.R_sii = setup.R_sii
        self.nabla_iiv = setup.nabla_iiv
        self.data = SimpleNamespace(phi_jg=setup.data.phi_jg,
                                    phit_jg=setup.data.phit_jg)
        self.xc_correction = SimpleNamespace(
            rgd=setup.xc_correction.rgd, Y_nL=setup.xc_correction.Y_nL,
            n_qg=setup.xc_correction.n_qg, nt_qg=setup.xc_correction.nt_qg,
            nc_g=setup.xc_correction.nc_g, nct_g=setup.xc_correction.nct_g,
            nc_corehole_g=setup.xc_correction.nc_corehole_g,
            B_pqL=setup.xc_correction.B_pqL, e_xc0=setup.xc_correction.e_xc0)
        self.hubbard_u = setup.hubbard_u
